# coded by lagcleaner
# email: lagcleaner@gmail.com

from io import BytesIO
from typing import Any
import socket
import tfprotocol_client.connection.socks_prox as socks
from tfprotocol_client.misc.constants import DFLT_HEADER_SIZE, DFLT_MAX_BUFFER_SIZE
from tfprotocol_client.misc.guard_exception import (
    on_except,
    raise_from,
)
from tfprotocol_client.misc.timeout_func import TimeLimitExpired, timelimit
from tfprotocol_client.models.exceptions import ErrorCode, TfException

from tfprotocol_client.models.proxy_options import ProxyOptions
from tfprotocol_client.models.status_info import StatusInfo


class SocketClient:
    """Socket Client, handles proxy and basic send and receive. """

    def __init__(
        self,
        address: str = 'localhost',
        port: int = 1234,
        proxy_options: ProxyOptions = None,
        max_buffer_size: int = DFLT_MAX_BUFFER_SIZE,
        header_size: int = DFLT_HEADER_SIZE,
        verbosity_mode: bool = False,
    ) -> None:
        self._socket: socks.socksocket = socks.socksocket(
            socket.AF_INET, socket.SOCK_STREAM
        )
        if proxy_options is not None:
            self._socket.set_proxy(
                proxy_options.proxy_type,
                proxy_options.address,
                proxy_options.port,
                username=proxy_options.username,
                password=proxy_options.password,
            )
        self._proxy_options = proxy_options
        self.address: str = address
        self.port: int = port
        self.header_size: int = header_size
        self.max_buffer_size: int = max_buffer_size
        self._is_connect: bool = False
        self.verbosity_mode = verbosity_mode

    def is_connect(self):
        return self._is_connect

    @property
    def socket(self) -> socket.socket:
        return self._socket

    @socket.setter
    def set_socket(self, new_socket: socks.socksocket):
        self._socket = new_socket

    @on_except(
        dflt_handler=lambda _: StatusInfo.parse("DISCONNECTED 0 connection time out"),
        **{
            AttributeError.__name__: lambda _: StatusInfo.parse(
                "DISCONNECTED 0 null pointer exception"
            ),
            TimeLimitExpired.__name__: lambda _: StatusInfo.parse(
                "DISCONNECTED 0 time out dns"
            ),
        },
    )
    def start_connection(self, dns_resolution_timeout: int, timeout: int) -> Any:
        """Method to initiate the connection with the server via socket.

        Args:
            `dns_resolution_timeout` (int): dns timeout expressed in seconds
            `timeout` (int): timeout to connect expresed in seconds

        Returns:
            StatusInfo: status information resulting from connection attempt
        """
        ip = timelimit(
            dns_resolution_timeout,
            (lambda self: socket.gethostbyname(self.address)),
            args=(self,),
        )
        self._socket.settimeout(timeout)
        self._socket.connect((ip, self.port))
        self._is_connect = True
        return StatusInfo.parse("OK")

    @on_except(dflt_handler=raise_from)
    def stop_connection(self):
        """Close socket connection to the server"""
        if self._socket is not None:
            try:
                self._socket.close()
            except Exception:  # pylint: disable=broad-except
                pass
        self._is_connect = False

    def _send(self, rawmessage: bytes) -> int:
        """Send a bunch of bytes through the socket to the server.

        Args:
            `rawmessage` (bytes): Ready to send message bytes.

        Raises:
            TfException: Socket write failed.

        Returns:
            int: how many bytes has been sended.
        """
        try:
            if rawmessage:
                return self.socket.send(rawmessage)
            return 0
        except Exception as e:
            raise TfException(
                message='Cannot succesfully write to socket',
                code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
                exception=e,
            )

    def _recv(self, size: int) -> bytes:
        """Receive a bunch of bytes from the socket connection

        Args:
            `size` (int): The size of the message to read.

        Raises:
            TfException: Memory limit reached.
            TfException: Socket sudently closed.
            TfException: Invalid size.

        Returns:
            bytes: message received.
        """
        if size < 0:
            raise TfException(message='Invalid byte size message to recieve.')
        binary_message = BytesIO()
        bytes_received = 0
        raw_chunk = bytearray(min(self.max_buffer_size, size - bytes_received))
        while bytes_received < size:
            try:
                # read chunk
                read = 0
                read = self.socket.recv_into(
                    raw_chunk, min(len(raw_chunk), size - bytes_received)
                )

                # save chunk
                bytes_received += read
                binary_message.write(raw_chunk[:read])
            except OSError as e:
                raise TfException(exception=e, message="Cannot read from socket ...")
            except MemoryError as excpt:
                # header_size is not going to reach this line
                # cause is an integer of up to 8 bytes only
                raise TfException(
                    exception=excpt,
                    code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
                    message=f'Heap space not enough server answer with a very \
                        high header number\nheader:{size}',
                )
        return binary_message.getvalue()

    def exception_guard(self):
        if self.socket is None or not self.is_connect():
            raise TfException(
                message="Socket is closed or not connected",
                code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
            )

    def __del__(self):
        self.stop_connection()



# coded by lagcleaner
# email: lagcleaner@gmail.com

from tfprotocol_client.connection.protocol_client import ProtocolClient
from tfprotocol_client.misc.constants import LONG_SIZE
from tfprotocol_client.models.putget_commands import PutGetCommandEnum


class CodesSenderRecvr:
    """Codes sender and receiver easy to use class. To handle signals of PUT and GET
    commands.
    """

    def __init__(self, client: ProtocolClient) -> None:
        """Codes utility initialization.

        Args:
            client (ProtocolClient): Protocol client.
        """
        self._client: ProtocolClient = client
        self.recveing_signal: bool = False
        self.sending_signal: bool = False
        self.block: bool = False
        self._last_command = PutGetCommandEnum.HPFEND.value
        self.last_header: int = 0

    def send_put(self, command: int):
        """Send a code to the server codes can only be codes described at XSAceConsts
        class.

        Args:
            command (int[64-bit]): The code to be send to server.
        """
        if not self.block:
            self.recveing_signal = True
            self._last_command = command
            self._client.just_send(command, size=LONG_SIZE, signed=True)

    def send_get(self, command: int):
        """Send a code to the server codes can only be codes described at XSAceConsts
        class.

        Args:
            command (int[64-bit]): The code to be send to server.
        """
        if not self.block:
            self.sending_signal = True
            self._last_command = command
            self._client.just_send(command, size=LONG_SIZE, signed=True)

    @property
    def last_command(self):
        """Return the last command sent."""
        return self._last_command



# coded by lagcleaner
# email: lagcleaner@gmail.com

class TransferStatus:
    """Transfer status"""

    def __init__(self) -> None:
        self.dummy = None
        self.command = None
        self.dummy_state = None

    def __str__(self) -> str:
        return f'TransferStatus<{self.dummy}, {self.command}, {self.dummy_state}>'



# coded by lagcleaner
# email: lagcleaner@gmail.com

from enum import Enum


class PutGetCommandEnum(Enum):
    """Commands used to put and get commands optimization."""

    HPFEND = 0
    HPFSTOP = -1
    HPFCANCEL = -2
    HPFCONT = -3
    HPFFIN = -127



# coded by lagcleaner
# email: lagcleaner@gmail.com

from typing import Optional, Union
from multipledispatch import dispatch
from tfprotocol_client.connection.client import SocketClient
from tfprotocol_client.misc.constants import (
    DFLT_MAX_BUFFER_SIZE,
    ENDIANESS_NAME,
    INT_SIZE,
)
from tfprotocol_client.misc.build_utils import MessageUtils
from tfprotocol_client.models.exceptions import TfException
from tfprotocol_client.models.message import TfProtocolMessage
from tfprotocol_client.models.proxy_options import ProxyOptions
from tfprotocol_client.models.status_info import StatusInfo
from tfprotocol_client.security.cryptography import Xor


class ProtocolClient(SocketClient):
    """ Protocol Client to handle, connections and make the interface
    easy to use.

    SocketClient: Super class
    """

    def __init__(
        self,
        address: str = 'localhost',
        port: int = 1234,
        proxy_options: Optional[ProxyOptions] = None,
        max_buffer_size = DFLT_MAX_BUFFER_SIZE,
        verbosity_mode: bool = False,
    ) -> None:
        super().__init__(
            address=address,
            port=port,
            proxy_options=proxy_options,
            max_buffer_size=max_buffer_size,
            verbosity_mode=verbosity_mode,
        )
        self.builder_utils: MessageUtils = MessageUtils(max_buffer_size=max_buffer_size)
        self.xor_input = None
        self.xor_output = None
        self._session_key = None

    def get_sessionkey(self):
        return self._session_key

    def set_sessionkey(self, session_key: bytes):
        if session_key is not None:
            self.xor_input = Xor(session_key)
            self.xor_output = Xor(session_key)
        self._session_key = session_key

    session_key = property(fget=get_sessionkey, fset=set_sessionkey)

    def _decrypt(self, payload: bytes) -> bytes:
        if self.xor_input is not None:
            payload = self.xor_input.decrypt(payload)
        return payload

    def _encrypt(self, payload: bytes) -> bytes:
        if self.xor_output is not None:
            payload = self.xor_output.encrypt(payload)
        return payload

    def just_recv_int(self, size: int = INT_SIZE, signed=False) -> int:
        # RECEIVE, DECRYPT AND DECODE INTEGER
        decrypted_data = self.just_recv(size)
        return MessageUtils.decode_int(decrypted_data, signed=signed)

    def just_recv_str(self, size: int) -> str:
        # RECEIVE, DECRYPT AND DECODE TEXT-STRING
        decrypted_data = self.just_recv(size)
        return MessageUtils.decode_str(decrypted_data)

    def just_recv(self, size: int = None) -> bytes:
        if not size:
            raise TfException(message="Bytes to receive not specified ...")
        # RECEIVE AND DECRYPT CHUNK
        received_header = self._recv(size)
        decrypted_header = self._decrypt(received_header)
        return decrypted_header

    @dispatch((int, str, bytes, bool))
    def just_send(self, message: int, size=INT_SIZE, signed=False, **_):
        self.exception_guard()

        # BUILD
        encoded_message = MessageUtils.encode_value(message, size=size, signed=signed)

        # ENCRYPT
        encrypted_message: bytes = self._encrypt(encoded_message)

        # SEND
        self._send(encrypted_message)

    # pylint: disable=function-redefined
    @dispatch(TfProtocolMessage)
    def just_send(self, message: TfProtocolMessage, **_):
        self.exception_guard()

        # BUILD
        _, encoded_message = message

        # ENCRYPT
        encrypted_message: bytes = self._encrypt(encoded_message)

        # SEND
        self._send(encrypted_message)

    @dispatch(TfProtocolMessage)
    def send(self, message: TfProtocolMessage, **_):
        self.exception_guard()
        # BUILD
        header, encoded_message = message
        if self.verbosity_mode:
            print(
                f'CLIENT: {int.from_bytes(header, byteorder=ENDIANESS_NAME)} {encoded_message}'
            )

        # ENCRYPT
        encrypted_header: bytes = self._encrypt(header)
        encrypted_message: bytes = self._encrypt(encoded_message)

        # SEND
        self._send(encrypted_header)
        self._send(encrypted_message)

    # pylint: disable=function-redefined
    @dispatch((str, bytes))
    def send(
        self,
        message: Union[str, bytes],
        custom_header: Union[str, bytes, int, bool, None] = None,
        header_size: int = None,
        **_,
    ):
        self.send(
            TfProtocolMessage(
                message, custom_header=custom_header, header_size=header_size,
            )
        )

    @dispatch()
    def recv(self, header_size=None, header_signed=True) -> StatusInfo:
        self.exception_guard()
        header_size = header_size if header_size > 0 else self.header_size
        # RECEIVE, DECRYPT AND DECODE HEADER
        received_header = self._recv(header_size)
        decrypted_header = self._decrypt(received_header)
        decoded_header = MessageUtils.decode_int(decrypted_header, signed=header_signed)
        # print(f'SERVER: Header({decoded_header})')

        # RECEIVE AND DECRYPT BODY
        received_body = self._recv(decoded_header)
        decrypted_body = self._decrypt(received_body)
        status = StatusInfo.build_status(decoded_header, decrypted_body)
        if self.verbosity_mode:
            print(f'SERVER: {decoded_header} {status}')
        return status

    @dispatch(TfProtocolMessage)
    def translate(
        self, message: TfProtocolMessage, recv_header_signed=True, **_
    ) -> StatusInfo:
        self.exception_guard()
        self.send(message)
        return self.recv(
            header_size=message.header_size, header_signed=recv_header_signed
        )

    # pylint: disable=function-redefined
    @dispatch((str, bytes))
    def translate(
        self,
        message: Union[str, bytes],
        custom_header: Union[str, bytes, int, bool, None] = None,
        header_size: int = None,
        recv_header_signed=True,
        **_,
    ) -> StatusInfo:
        return self.translate(
            TfProtocolMessage(
                message, custom_header=custom_header, header_size=header_size
            ),
            recv_header_signed=recv_header_signed,
        )



# coded by lagcleaner
# email: lagcleaner@gmail.com

from enum import Enum
from typing import Optional, Union
from tfprotocol_client.misc.constants import STRING_ENCODING
from tfprotocol_client.misc.status_server_code import StatusServerCode
from tfprotocol_client.models.status_info import StatusInfo


class ErrorCode(Enum):
    """Error Code for `TfExceptions`"""

    ON_WRITE_OR_RECEIVE_TO_SOCKET = 0
    ILLEGAL_ARGUMENTS = 1
    UNHANDLED_EXCEPTION = 2
    CAN_PUT = 3


class TfException(Exception):
    """Transfer Protocol Exception"""

    def __init__(
        self,
        *args: tuple,
        exception: Optional[Exception] = None,
        message: Optional[str] = None,
        status_info: Optional[StatusInfo] = None,
        status_server_code: Optional[StatusServerCode] = None,
        code: Union[int, ErrorCode, None] = None,
    ) -> None:
        super().__init__(*args, message)
        if status_info:
            self.status_info: StatusInfo = status_info
        else:
            self.status_info = StatusInfo(status=StatusServerCode.FAILED)
        if exception:
            self.original_exception: Exception = exception
            self.status_info.code = hash(exception)
            self.status_info.message = str(exception)
        if status_server_code:
            self.status_info.status = status_server_code
        if message:
            self.status_info.message = message
        if code:
            self.status_info.code = code.value if isinstance(code, ErrorCode) else code

    @property
    def payload(self):
        return (
            self.status_info.payload
            if self.status_info.payload is not None
            else self.status_info.message.encode(STRING_ENCODING)
        )

    def __str__(self) -> str:
        return f'''\n--------------------------------------------------
{self.status_info.status}
{self.status_info.code}
{self.status_info.message}
--------------------------------------------------\n'''



# coded by lagcleaner
# email: lagcleaner@gmail.com

import hashlib
from typing import Union


def sha256_for(payload: Union[bytes, str]):
    '''Extract a unique SHA256 hash from a payload'''
    binary_data = payload.encode("utf8") if isinstance(payload, str) else payload
    return hashlib.sha256(binary_data).digest()


def hexstr(data: bytes) -> str:
    return '0x' + data.hex()



# coded by lagcleaner
# email: lagcleaner@gmail.com

from io import BytesIO
from struct import unpack
from typing import Union
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from Crypto.Cipher import PKCS1_OAEP
from Crypto.Hash import SHA1


class CryptographyUtils:
    """Cryptography utils to handle public key and generate random bytes."""

    @staticmethod
    def rsa_encrypt(payload: bytes, public_key: str) -> bytes:
        recipient_key = RSA.import_key(public_key)
        # Encrypt payload with the public RSA key
        cipher_rsa = PKCS1_OAEP.new(recipient_key, hashAlgo=SHA1)
        enc_payload = cipher_rsa.encrypt(payload)
        return enc_payload

    get_random_bytes = staticmethod(get_random_bytes)


class Xor:
    """Xor class to en/decrypt messages."""

    SESSION_KEY = None

    def __init__(self, key: Union[bytes, bytearray] = None) -> None:
        if key is not None and (isinstance(key, bytes) or isinstance(key, bytearray)):
            Xor.SESSION_KEY = bytearray(key)
        self._session_key = Xor.SESSION_KEY[:]
        self._key = Xor.SESSION_KEY[:]
        self._seed = self._get_new_seed()

    def _get_new_seed(self) -> int:
        buff = BytesIO(self._key)
        (new_seed,) = unpack('<q', buff.read(8))
        return new_seed

    def get_seed(self) -> int:
        return self._seed

    def encrypt(self, payload: bytes) -> bytearray:
        mut_payload = bytearray(payload)
        for i, _ in enumerate(mut_payload):
            mut_payload[i] ^= self._key[i % len(self._key)]
            # Pack data to send
            mut_payload[i] = (mut_payload[i] + (self._seed >> 56 & 0xFF)) % 256
            # Change seed and encryption key
            self._seed = self._seed * (self._seed >> 8 & 0xFFFFFFFF) + (
                self._seed >> 40 & 0xFFFF
            )
            if self._seed == 0:
                self._seed = self._get_new_seed()
            self._key[i % len(self._key)] = self._seed % 256

        return mut_payload

    def decrypt(self, payload: bytes) -> bytearray:
        mut_payload = bytearray(payload)
        for i, _ in enumerate(mut_payload):
            # Unpack received data
            mut_payload[i] = (256 + mut_payload[i] - (self._seed >> 56 & 0xFF)) % 256
            # Decrypt received data
            mut_payload[i] ^= self._key[i % len(self._key)]
            self._seed = self._seed * (self._seed >> 8 & 0xFFFFFFFF) + (
                self._seed >> 40 & 0xFFFF
            )
            if self._seed == 0:
                self._seed = self._get_new_seed()

            self._key[i % len(self._key)] = self._seed % 256

        return mut_payload



# coded by lagcleaner
# email: lagcleaner@gmail.com

class ProxyOptions:
    """Proxy Options model to contain the options in case of proxy usage.
    """

    def __init__(
        self, proxy_type, address: str, port: int, username: str, password: str,
    ) -> None:
        """Proxy Options container

        Args:
            `proxy_type` (proxy type): The type of the proxy to be used. Three types are
            supported: PROXY_TYPE_SOCKS4 (including socks4a), PROXY_TYPE_SOCKS5 and PROXY_TYPE_HTTP
            `address` (str): The address of the server (IP or DNS).
            `port` (int): The port of the server. Defaults to 1080 for SOCKS servers and 8080 for
            HTTP proxy servers.
            `username` (str): Username to authenticate with to the server. The default is no
            authentication. (Optional)
            `password` (str): Password to authenticate with to the server. Only relevant when
            username is also provided. (Optional)
        """
        self.proxy_type = proxy_type
        self.port = port
        self.address = address
        self.username = username.encode() if username else None
        self.password = password.encode() if password else None



# coded by lagcleaner
# email: lagcleaner@gmail.com

from typing import Optional
from tfprotocol_client.misc.parse_utils import (
    separate_status_b,
    separate_status_codenumber_b,
    tryparse_int,
)
from tfprotocol_client.misc.constants import STRING_ENCODING
from tfprotocol_client.misc.status_server_code import StatusServerCode


class StatusInfo:
    """Status Info type from server responses."""

    def __init__(
        self,
        status: StatusServerCode = None,
        opcode: int = None,
        sz: int = None,  # pylint: disable=invalid-name
        payload: bytes = b'',
        code: int = 0,
        message: str = '',
    ):
        self.status: StatusServerCode = status
        self.opcode: int = opcode
        self.payload: bytes = payload if payload else b''
        self.code: int = code
        self.sz: int = sz  # pylint: disable=invalid-name
        self.message: str = message

    @staticmethod
    def parse(
        rawmessage: Optional[str] = None, payload: Optional[bytes] = None
    ) -> 'StatusInfo':
        status_info = None
        if rawmessage is not None:
            values = rawmessage.split(' ')
            if len(values) == 1:
                status_info = StatusInfo(StatusServerCode.from_str(values[0]), 0, '')
            else:
                status_info = StatusInfo(
                    StatusServerCode.from_str(values[0]),
                    int(values[1]) if values[1].isdigit() else None,
                    ' '.join(values[2:]),
                )
        if payload is not None:
            status_info.payload = payload
        return status_info if status_info else StatusInfo()

    @staticmethod
    def build_status(
        header: int, message: bytes, parse_code: bool = True
    ) -> 'StatusInfo':
        code: int = header
        # msg_str_full: str = str(message, encoding=STRING_ENCODING)
        status, msg = separate_status_b(message)

        if status is None:
            # ? <msg>
            status, msg = StatusServerCode.UNKNOWN, message
        elif (status is not StatusServerCode.FAILED) or not parse_code:
            # ? <status> [<msg>]
            # msg = msg.replace(
            #     str.encode(status.name, encoding=STRING_ENCODING), b'', 1
            # ).strip()
            code = status.value
        else:
            # ? FAILED <str_code> : <msg>
            str_code, msg = separate_status_codenumber_b(msg)
            msg = msg.strip().replace(b': ', b'', 1)
            code = tryparse_int(str_code, dflt_value=status.value)

        return StatusInfo(
            status,
            code=code,
            message=str(
                msg[:1024] + b'...' if len(msg) > 1024 else msg,
                encoding=STRING_ENCODING,
            ),
            payload=msg,
        )

    def __str__(self):
        return f'StatusInfo[{self.status.name}]<{self.code}, "{self.message}">'



# coded by lagcleaner
# email: lagcleaner@gmail.com

from io import BytesIO
import sys
from platform import system as current_operating_system
import time
import socket
from threading import Thread
from typing import Callable, Optional
import tfprotocol_client.connection.socks_prox as socks
from tfprotocol_client.connection.protocol_client import ProtocolClient
from tfprotocol_client.misc.build_utils import MessageUtils
from tfprotocol_client.misc.constants import BYTE_SIZE, EMPTY_HANDLER
from tfprotocol_client.models.exceptions import TfException
from tfprotocol_client.models.keepalive_options import (
    KeepAliveMechanismType,
    KeepAliveOptions,
)
from tfprotocol_client.models.proxy_options import ProxyOptions
from tfprotocol_client.tfprotocol_keepalive import TfProtocolKeepAliveWrapper


class KeepAliveThread(Thread):
    """Keep Alive mechanism threaded."""

    def __init__(
        self,
        proto_client: ProtocolClient,
        options: KeepAliveOptions,
        proxy_options: ProxyOptions = None,
        on_connection_closed: Callable[[], None] = EMPTY_HANDLER,
    ) -> None:
        super().__init__(name='keepalive_t')
        self.is_active = False
        self._keepalive_mechanism: KeepAliveMechanismType = options.keepalive_mechanism
        self._proxy_options: ProxyOptions = proxy_options
        self._proto_client: ProtocolClient = proto_client
        self._idle: int = options.idle
        self._timeout: int = options.timeout
        self._max_tries: int = options.max_tries
        self._counter: int = 0
        self._prockey: str = ''
        self._on_connection_closed: Callable[[],None] = on_connection_closed

        try:
            self._datagram_socket: Optional[socks.socksocket] = socks.socksocket(
                socket.AF_INET, socket.SOCK_DGRAM
            )
            if proxy_options is not None:
                self._datagram_socket.set_proxy(
                    proxy_options.proxy_type,
                    proxy_options.address,
                    proxy_options.port,
                    username=proxy_options.username,
                    password=proxy_options.password,
                )
            self._datagram_socket.settimeout(options.timeout)
            success = False
            if options.keepalive_mechanism is KeepAliveMechanismType.TCP_NATIVE:
                if current_operating_system() == 'Windows':
                    set_keepalive_windows(
                        self._proto_client.socket,
                        self._idle,
                        self._timeout,
                        self._max_tries,
                    )
                elif current_operating_system() == 'Linux':
                    set_keepalive_linux(
                        self._proto_client.socket,
                        self._idle,
                        self._timeout,
                        self._max_tries,
                    )
                elif current_operating_system() == 'MacOS':
                    set_keepalive_linux(
                        self._proto_client.socket,
                        self._idle,
                        self._timeout,
                        self._max_tries,
                    )

            if options.keepalive_mechanism is KeepAliveMechanismType.UDP_PROCHECK:
                dflt_timout = self._proto_client.socket.gettimeout()
                self._proto_client.socket.settimeout(options.timeout)
                proto_wrapper = TfProtocolKeepAliveWrapper(self._proto_client)
                for _ in range(options.max_tries):
                    self._prockey = proto_wrapper.prockey_command()
                    self.is_active = proto_wrapper.keepalive_command(
                        True, 1, options.idle, options.max_tries
                    )
                    if not self._prockey or not self.is_active:
                        continue
                    self._proto_client.socket.settimeout(dflt_timout)
                    success = True
                    break
                if not success:
                    self.stop_connection()
        except IOError as e:
            raise TfException(exception=e)

    @property
    def client(self):
        return self._proto_client

    @property
    def udp_sock(self):
        return self._datagram_socket

    @property
    def addrs(self):
        return self._proto_client.address, self._proto_client.port

    def stop_connection(self):
        self._datagram_socket.close()
        self.client.stop_connection()
        self.is_active = False
        if self._on_connection_closed:
            self._on_connection_closed()
        return self.client.is_connect()

    def _stop(self):
        self.stop_connection()
        super()._stop()

    def udp_hostcheck(self):
        print(f'KA: udp_hostcheck (tries: {self._counter})')
        try:
            data_to_send = bytes((0,))
            _ = self.udp_sock.sendto(data_to_send, self.addrs)
            data_recvd, _ = self.udp_sock.recvfrom(1)
            print(f'KA: udp_hostcheck (recv: {data_recvd})')
            if data_recvd[0] == 1:
                self._counter = 0
        except IOError as e:
            print('UDP_SOCKET-IOError: ', e.strerror, e.args)
        except e:
            print('UDP_SOCKET-Error: ', e)
        finally:
            if self._counter == self._max_tries:
                self.stop_connection()
            else:
                try:
                    time.sleep(self._idle)
                except InterruptedError:
                    pass
            self._counter += 1

    def udp_procheck(self):
        print('KA: udp_procheck')
        try:
            payload = BytesIO()
            payload.write(MessageUtils.encode_value(1, size=BYTE_SIZE))
            payload.write(MessageUtils.encode_value(self._prockey))
            print('KA: udp_procheck - send', payload.getvalue())
            _ = self.udp_sock.sendto(payload.getvalue(), self.addrs)

            data_recvd, _ = self.udp_sock.recvfrom(1)
            print('KA: udp_procheck - recv', data_recvd)
            if data_recvd[0] == 0:
                self.stop_connection()
                return
            if data_recvd[0] == 1:
                self._counter = 0
        except IOError:
            pass
        finally:
            self._counter += 1
            if self._counter == self._max_tries:
                self.stop_connection()
            else:
                try:
                    time.sleep(self._idle)
                except InterruptedError:
                    pass

    def run(self) -> None:
        # return super().run()
        print('KA: run')
        self.is_active = True
        while self.is_active and self.client.is_connect():
            if self._keepalive_mechanism is KeepAliveMechanismType.UDP_HOSTCHECK:
                self.udp_hostcheck()
            elif self._keepalive_mechanism is KeepAliveMechanismType.UDP_PROCHECK:
                self.udp_procheck()
            elif self._keepalive_mechanism is KeepAliveMechanismType.TCP_NATIVE:
                break
            else:
                sys.stderr.write(
                    f'Unhandled exception: {self._keepalive_mechanism} implementation not found.'
                )
                sys.exit(1)


def set_keepalive_linux(sock, after_idle_sec=3600, interval_sec=3, max_fails=5):
    """Set TCP keepalive on an open socket.

    It activates after 3600 seconds (after_idle_sec) of idleness,
    then sends a keepalive ping once every 3 seconds (interval_sec),
    and closes the connection after 5 failed ping (max_fails), or 15 seconds
    """
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)


def set_keepalive_windows(sock, after_idle_sec=3600, interval_sec=3, max_fails=5):
    """Set TCP keepalive on an open socket.

    It activates after 3600 seconds (after_idle_sec) of idleness,
    then sends a keepalive ping once every 3 seconds (interval_sec),
    and closes the connection after 5 failed ping (max_fails), or 15 seconds
    """
    sock.ioctl(
        socket.SIO_KEEPALIVE_VALS, (1, after_idle_sec * 1000, interval_sec * 1000),
    )


def set_keepalive_mac(sock, after_idle_sec=3600, interval_sec=3, max_fails=5):
    """Set TCP keepalive on an open socket.

    It activates after 3600 seconds (after_idle_sec) of idleness,
    then sends a keepalive ping once every 3 seconds (interval_sec),
    and closes the connection after 5 failed ping (max_fails), or 15 seconds
    """
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, interval_sec)
    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, max_fails)



# coded by lagcleaner
# email: lagcleaner@gmail.com

from typing import Callable
from tfprotocol_client.misc.status_server_code import StatusServerCode

from tfprotocol_client.models.status_info import StatusInfo


class TCPTimeoutOptions:
    """TCP timeout options. """

    def __init__(
        self,
        connect_timeout: int = 120 * 1000,
        connect_retry: int = 3,
        dns_resolution_timeout: int = 20 * 1000,
        status_server_callback: Callable[[StatusInfo], None] = None,
    ) -> None:
        self._connect_retry: int = connect_retry
        self._connect_timeout: int = connect_timeout
        self._dns_resolution_timeout: int = dns_resolution_timeout
        self.status_server_callback = (
            (lambda _: None)
            if status_server_callback is None
            else status_server_callback
        )

    @property
    def connect_timeout(self) -> int:
        return self._connect_timeout

    @connect_timeout.setter
    def set_connect_timeout(self, value: int):
        if value < 0:
            self.status_server_callback(
                StatusInfo(
                    status=StatusServerCode.FAILED,
                    opcode=-1,
                    message='connect_timeout has to be a non-negative number',
                )
            )
        else:
            self._connect_timeout = value

    @property
    def connect_retry(self) -> int:
        return self._connect_retry

    @connect_retry.setter
    def set_connect_retry(self, value: int):
        if value < 0:
            self.status_server_callback(
                StatusInfo(
                    status=StatusServerCode.FAILED,
                    opcode=-1,
                    message='connect_retry has to be a non-negative number',
                )
            )
        else:
            self._connect_retry = value

    @property
    def dns_resolution_timeout(self) -> int:
        return self._dns_resolution_timeout

    @dns_resolution_timeout.setter
    def set_dns_resolution_timeout(self, value: int):
        self._dns_resolution_timeout = value



# coded by lagcleaner
# email: lagcleaner@gmail.com

from enum import Enum


class KeepAliveMechanismType(Enum):
    """Keep alive mechanism types
    """

    UDP_HOSTCHECK = 0
    UDP_PROCHECK = 1
    TCP_NATIVE = 2


class KeepAliveOptions:
    """Keep Alive model to contain the options in case of keep alive.
    """

    def __init__(
        self,
        keepalive_mechanism: KeepAliveMechanismType,
        idle: int,
        timeout: int,
        max_tries: int,
    ) -> None:
        super().__init__()
        self.keepalive_mechanism: KeepAliveMechanismType = keepalive_mechanism
        self.idle: int = idle
        self.timeout: int = timeout
        self.max_tries: int = max_tries



# coded by lagcleaner
# email: lagcleaner@gmail.com

from threading import Thread


class TimeLimitExpired(Exception):
    pass


def timelimit(timeout, func, args=(), kwargs={}):
    """ Run func with the given timeout. If func didn't finish running
        within the timeout, raise TimeLimitExpired
    """

    class FuncThread(Thread):
        def __init__(self):
            Thread.__init__(self)
            self.result = None

        def run(self):
            self.result = func(*args, **kwargs)

        def stop(self):
            if self.is_alive():
                self._stop()

    it = FuncThread()
    it.start()
    it.join(timeout)
    if it.is_alive():
        it.stop()
        raise TimeLimitExpired()
    else:
        return it.result



# coded by lagcleaner
# email: lagcleaner@gmail.com

from datetime import datetime
from enum import Enum
import struct
from typing import Final, Tuple, Union
from tfprotocol_client.misc.constants import ENDIANESS
from tfprotocol_client.models.exceptions import TfException


class FileStatTypeEnum(Enum):
    """Describes a type of File.
    """

    DIR = 0
    FILE = 1
    UNKNOWN = 2

    @staticmethod
    def from_char(bchar: str, dflt=None):
        try:
            for enum_v in FileStatTypeEnum:
                if enum_v.name.startswith(bchar):
                    return enum_v
            return dflt
        except (KeyError, ValueError):
            return dflt


class FileStat:
    """File Stat object, used to contain file info.
    """

    def __init__(
        self,
        filestat_type: Union[int, str, FileStatTypeEnum],
        size: int,
        last_access: int,
        last_modification: int,
    ) -> None:
        self.type: Final[Union[FileStatTypeEnum, int]] = filestat_type
        self.size: Final[int] = size
        self.last_access: Final[int] = last_access
        self.last_modification: Final[int] = last_modification

    @property
    def last_modification_date(self) -> datetime:
        if self.last_modification:
            return datetime.fromtimestamp(self.last_modification)
        return None

    @property
    def last_access_date(self) -> datetime:
        if self.last_access:
            return datetime.fromtimestamp(self.last_access)
        return None

    def __str__(self) -> str:
        return (
            f'FileStat(type={self.type}, '
            f'size={self.size}, '
            f'last_access={self.last_access}, '
            f'last_mod={self.last_modification})'
        )

    @staticmethod
    def build_from_structure(fstatstruct: bytes) -> Tuple[int, 'FileStat']:
        """Build a FileStat object from fstatstruct structure."""
        if len(fstatstruct) != 26:
            raise TfException(message='Invalid format of `fstathdr`...')
        code, _type, size, atime, mtime = struct.unpack(
            f'{ENDIANESS}bbQQQ', fstatstruct
        )
        # TODO: _TYPE PENDING TO VERIFICATION
        return code, FileStat(_type, size, atime, mtime)



# coded by lagcleaner
# email: lagcleaner@gmail.com

from functools import wraps
from typing import Any, Callable, Dict, Union
from tfprotocol_client.models.exceptions import ErrorCode, TfException
from tfprotocol_client.models.status_info import StatusInfo

ExceptionHandler = Callable[[Exception], Any]


def raise_from(excp: Exception):
    """Raise TfException from regular exceptions.

    Args:
        excp (Exception): Regular exceptions raised to feed TfExceptions.

    Raises:
        TfException: Custom exception
    """
    raise TfException(exception=excp)


def get_raise_exceptions_handler(
    message: str = None,
    code: Union[int, ErrorCode, None] = None,
    status_info: StatusInfo = None,
):
    """Retrieves a exception handler.

    Args:
        message (str, optional): Exception message. Defaults to None.
        code (Union[int, ErrorCode, None], optional): Error code. Defaults to None.
        status_info (StatusInfo, optional): Status info. Defaults to None.
    """

    def _raise_from(excp: Exception):
        raise TfException(
            exception=excp, message=message, code=code, status_info=status_info,
        )

    return _raise_from


def on_except(
    dflt_handler: ExceptionHandler = None, **handlers: Dict[str, ExceptionHandler]
):
    """Handle exceptions or wraps exceptions and re-raise.

    Args:
        `dflt_handler` (ExceptionHandler): Default handler for exceptions.
        `**handlers` (Dict[str, ExceptionHandler]): Handlers dictionary exception name
        to handler function.
    """

    def _on_except(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as excp:
                handler = handlers.get(excp.__class__.__name__, dflt_handler)
                if handler is None:
                    raise excp
                return handler(excp)

        return _wrapper

    return _on_except


def exception_raise_guard(func):
    @wraps(func)
    def _exception_raise_guard(*args, **kwargs):
        self = args[0]
        if self.socket is None or not self.is_connect():
            raise TfException(
                message="Socket is closed or not connected",
                code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
            )
        return func(*args, **kwargs)




# coded by lagcleaner
# email: lagcleaner@gmail.com

LONG_SIZE = 8
INT_SIZE = 4
SHORT_SIZE = 2
BYTE_SIZE = 1

DFLT_MAX_BUFFER_SIZE = 8 * 1024  # 512 * 1024
DFLT_HEADER_SIZE = INT_SIZE

# Key len interval in bytes
KEY_LEN_INTERVAL = (16, 40)


ENDIANESS = '>'
ENDIANESS_NAME = 'little' if ENDIANESS == '<' else 'big'
STRING_ENCODING = 'UTF-8'  # UTF-8,LATIN-1, ISO-8859-1

EMPTY_HANDLER = lambda *_: None
RESPONSE_LOGGER = lambda resp: print(f"RESPONSE: {resp}")

# Set identity and it's permissions for the directory.
SECFS_SETPERM = 1
# Remove identity and it’s permissions for the directory.
SECFS_REMPERM = 2
# Set Read permission for files in the directory.
SECFS_RFILE = 4
# Set Write -Create, Truncate, etc.- permission for files in the directory.
SECFS_WFILE = 8
# Set List permission for files in the directory.
SECFS_LDIR = 16
# Set Remove Directory permission for directories in the directory.
SECFS_RMDIR = 32
# Set Make Directory permission for directories in the directory.
SECFS_MKDIR = 64
# Set Delete permission for files in the directory.
SECFS_DFILE = 128
# Set STAT permission for files and directories in the directory
SECFS_STAT = 256
# Set File Updating Timestamps permission for files and directories in the directory.
SECFS_FDUPD = 512
# Set UNIX permission changing for files and directories in the directory.
SECFS_UXPERM = 1024
# Set List Directory Recursively permission for files and directories in the directory.
SECFS_LRDIR = 2048

SECFS_ALL_PERMISSIONS = (
    SECFS_SETPERM
    | SECFS_REMPERM
    | SECFS_RFILE
    | SECFS_DFILE
    | SECFS_LDIR
    | SECFS_RMDIR
    | SECFS_MKDIR
    | SECFS_DFILE
    | SECFS_STAT
    | SECFS_FDUPD
    | SECFS_UXPERM
    | SECFS_LRDIR
)



# coded by lagcleaner
# email: lagcleaner@gmail.com

import re
from typing import Tuple, Union

from tfprotocol_client.misc.constants import STRING_ENCODING
from tfprotocol_client.misc.status_server_code import StatusServerCode


def separate_status(string: str):
    res = re.split(r'\s+', string, maxsplit=1)
    # res_f = [r.strip() for r in res if r is not None and r.strip() != '']
    if len(res) > 1:
        return res[0], ''.join(res[1:])
    else:
        return res[0], ''


def separate_status_b(data: bytes) -> Tuple[StatusServerCode, bytes]:
    res = re.split(br'\s+', data, maxsplit=1)
    # res_f = [r.strip() for r in res if r is not None and r.strip() != '']
    status: StatusServerCode = StatusServerCode.from_str(
        str(res[0], encoding=STRING_ENCODING).upper()
    )

    if status is not None:
        return status, b''.join(res[1:]) if len(res) > 1 else b''
    return None, data


def separate_status_codenumber(string: str) -> Tuple[str, str]:
    res = re.split(r'([-+]?\d+\.\d+)|([-+]?\d+)', string.strip(), maxsplit=1)
    res_f = [r.strip() for r in res if r is not None and r.strip() != '']
    if len(res_f) > 1 and res_f[0].isdecimal():
        return res_f[0], ''.join(res_f[1:])
    else:
        return '', ''.join(res_f)


def separate_status_codenumber_b(data: bytes):
    res = re.split(br'([-+]?\d+\.\d+)|([-+]?\d+)', data.strip(), maxsplit=1)
    res_f = [r.strip() for r in res if r is not None and r.strip() != b'']
    if len(res_f) > 1 and isnumber(res_f[0]):
        return res[0], b''.join(res_f[1:])
    return b'', b''.join(res_f)


def isnumber(data: bytes):
    if data.isalnum():
        return True
    numb = data.split(b'.', maxsplit=1)
    return all(part.isalnum() for part in numb)


def tryparse_int(
    str_value: Union[str, bytes, bytearray], base: int = 10, dflt_value: int = 0
) -> int:
    try:
        if isinstance(str_value, (bytes, bytearray)):
            str_value = str(str_value, encoding=STRING_ENCODING)
        return int(str_value, base=base)
    except ValueError:
        return dflt_value



# coded by lagcleaner
# email: lagcleaner@gmail.com

from typing import Callable, Optional, Tuple, Union
from tfprotocol_client.connection.keep_alive_thread import KeepAliveThread
from tfprotocol_client.connection.protocol_client import ProtocolClient
from tfprotocol_client.misc.constants import (
    DFLT_MAX_BUFFER_SIZE,
    EMPTY_HANDLER,
    KEY_LEN_INTERVAL,
)
from tfprotocol_client.misc.status_server_code import StatusServerCode
from tfprotocol_client.models.keepalive_options import KeepAliveOptions
from tfprotocol_client.models.proxy_options import ProxyOptions
from tfprotocol_client.models.status_info import StatusInfo
from tfprotocol_client.models.tcptimeout_options import TCPTimeoutOptions
from tfprotocol_client.models.exceptions import ErrorCode, TfException
from tfprotocol_client.security.cryptography import CryptographyUtils


class TfProtocolSuper:
    """The top tfprotocol class, all other classes of tfprotocol extends
    from this one, callbacks extends from ISuperCallback, all classes
    which extends from this one must have in order to be used another
    class who extends from ISuperCallback, if doesn't then the user won't
    be able to instantiate the extended subclass.
    """

    def __init__(
        self,
        protocol_version: str,
        public_key: str,
        client_hash: Union[str, bytes],
        address: str,
        port: int,
        proxy: ProxyOptions = None,
        keylen: int = KEY_LEN_INTERVAL[0],
        channel_len: int = DFLT_MAX_BUFFER_SIZE,
        verbosity_mode: bool = False,
    ) -> None:
        """ Constructor for Transfer Protocol super class.

        Args:
            `protocol_version` (str): The desired version of the protocol.
            `public_key` (str): The previous shared rsa public key for
                the initial encryptation of the communication.
            `client_hash` (Union[str, bytes]): The hash to be used by
                the server in order to test the integrity of the communication.
            `address` (str): The ip/address where the protocol server is running.
            `port` (int): The TCP port where the protocol is listening.
            `proxy` (ProxyOptions, optional): The proxy address for
                connect through a proxy, proxy auth. Defaults to None.
            `keylen` (int, optional): The desired length for the session key
                used to encrypt communication whit the server. Defaults to KEY_LEN_INTERVAL[0].
            `channel_len` (int, optional): The length of the channel.
                Defaults to DFLT_MAX_BUFFER_SIZE.
            `verbosity_mode` (bool): Debug mode enabled for verbosity.
        """
        self.verbosity_mode = verbosity_mode
        self._protocol_version = protocol_version
        self._public_key = public_key
        self._client_hash = client_hash
        self._proto_client = ProtocolClient(
            address=address,
            port=port,
            proxy_options=proxy,
            verbosity_mode=verbosity_mode,
        )
        self._sleep_time_milisec = 3000
        self._len_channel = channel_len if channel_len is not None else 512 * 1024
        self._keybyteslen = (
            keylen if KEY_LEN_INTERVAL[0] <= keylen <= KEY_LEN_INTERVAL[1] else 16
        )
        self._address: Tuple[str, int] = address
        #
        self._tcp_timeout_options: Optional[TCPTimeoutOptions] = None

    def connect(
        self,
        keepalive_options: Optional[KeepAliveOptions] = None,
        on_response: Callable[[StatusInfo], None] = EMPTY_HANDLER,
        on_connect: Callable[['TfProtocolSuper'], None] = EMPTY_HANDLER,
    ):
        """Connect to the server with keep-alive mechanism enabled or disabled

        Args:
            `keepalive_options` (KeepAliveOptions): Keep alive options.
            `on_response` ((StatusInfo) -> None): The callback for the server responses in the
                connection process.
            `on_connect` ((TfProtocolSuper) -> None): The callback to retrieve the connected
                instance of tfprotocol after stablishing connection.
        """
        # TRY TO INITIATE CONNECTION
        final_status: StatusInfo = self._connect()
        if final_status.status is StatusServerCode.OK:
            if keepalive_options:
                # START KEEP ALIVE MECHANISM
                udp_keep_alive = KeepAliveThread(self.client, keepalive_options)
                udp_keep_alive.setDaemon(True)
                udp_keep_alive.start()
            on_connect(self)
            return True
        else:
            # FIXME: WHY THIS METHOD RAISE AN EXCEPTION AND THE STANDARD 'connect()' do not
            on_response(final_status)
            try:
                self._proto_client.stop_connection()
                raise TfException(
                    status_server_code=StatusServerCode.DISCONNECTED,
                    code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
                    message='Cannot successfully connect to the server',
                )
            except IOError as ex:
                raise TfException(exception=ex)

    def _connect(
        self, on_response: Callable[[StatusInfo], None] = EMPTY_HANDLER,
    ) -> StatusInfo:
        self._tcp_timeout_options = TCPTimeoutOptions(
            status_server_callback=on_response,
        )
        count = 0
        status: StatusInfo = None
        while count < self.tcp_timeout_options.connect_retry:
            count += 1
            # TRY TO STABLISH TCP CONNECTION
            status: StatusInfo = self._proto_client.start_connection(
                self.tcp_timeout_options.dns_resolution_timeout,
                self.tcp_timeout_options.connect_timeout,
            )
            if status.status == StatusServerCode.OK:
                break
        if status.status is not StatusServerCode.OK:
            # FAIL TO STABLISH CONNECTION
            return status

        # SEND PROTOCOL TO BE USED
        status: StatusInfo = self._proto_client.translate(self._protocol_version)
        if status.status is not StatusServerCode.OK:
            # INCOMPATIBLE PROTOCOLS
            return status

        # GENERATE AND SEND THE SESSION KEY
        session_key = CryptographyUtils.get_random_bytes(self._keybyteslen)
        enc_session_key = CryptographyUtils.rsa_encrypt(session_key, self._public_key)
        status: StatusInfo = self._proto_client.translate(enc_session_key)
        if status.status is not StatusServerCode.OK:
            return status

        # SAVE SESSION KEY
        self._proto_client.session_key = session_key

        # SEND CLIENT HASH
        status: StatusInfo = self._proto_client.translate(self._client_hash)
        if status.status is not StatusServerCode.OK:
            return status
        return StatusInfo(StatusServerCode.OK)

    def disconnect(self):
        """Disconect the protocol"""
        self.client.stop_connection()

    def get_len_channel(self) -> int:
        """Gets the len of the channel

        Returns:
            int: `len_channel`
        """
        return self._len_channel

    def set_len_channel(self, value: int):
        self._len_channel = value

    len_channel = property(fget=get_len_channel, fset=set_len_channel)

    @property
    def tcp_timeout_options(self) -> TCPTimeoutOptions:
        """ Gets options for connection attempts.

        Returns:
            TCPTimeoutOptions: `tcp_timeout_options`
        """
        return self._tcp_timeout_options

    @tcp_timeout_options.setter
    def set_tcp_timeout_options(self, value: TCPTimeoutOptions):
        self._tcp_timeout_options = value

    @property
    def client(self) -> ProtocolClient:
        return self._proto_client



# coded by lagcleaner
# email: lagcleaner@gmail.com

from io import BytesIO
from typing import Union
from multipledispatch import dispatch
from tfprotocol_client.misc.build_utils import MessageUtils
from tfprotocol_client.misc.constants import (
    DFLT_HEADER_SIZE,
    INT_SIZE,
)


class TfProtocolMessage:
    """Transfer Protocol Message object builder."""

    def __init__(
        self,
        *payloads,
        custom_header: Union[int, bytes, None] = None,
        header_signed: bool = True,
        trim_body: bool = False,
        separate_by_spaces: bool = True,
        header_size: int = DFLT_HEADER_SIZE,
    ) -> None:
        self.custom_header = MessageUtils.encode_value(
            custom_header, size=header_size, signed=header_signed
        )
        self.body_buffer = BytesIO()
        self.header_size = header_size if header_size else DFLT_HEADER_SIZE
        self.header_signed = header_signed
        for i, e in enumerate(payloads):
            if separate_by_spaces and i != 0:
                self.add(b' ')
            self.add(e)
        self.trim_body = trim_body

    @dispatch((str, bytes, bool))
    def add(self, payload: Union[str, bytes, bool], **_):
        self.body_buffer.write(MessageUtils.encode_value(payload))
        return self

    # pylint: disable=function-redefined
    @dispatch(int)
    def add(self, payload: int, size: int = INT_SIZE, signed=False, **_):
        self.body_buffer.write(
            MessageUtils.encode_value(payload, size=size, signed=signed)
        )
        return self

    @property
    def header(self) -> bytes:
        if self.custom_header is not None:
            header = MessageUtils.encode_value(self.custom_header)
        else:
            header = MessageUtils.encode_value(
                len(self.payload), size=self.header_size, signed=self.header_signed
            )
        return header

    @property
    def payload(self) -> bytes:
        if self.trim_body:
            return self.body_buffer.getvalue().strip(b' ')
        return self.body_buffer.getvalue()

    def __iter__(self):
        yield self.header
        yield self.payload



# coded by lagcleaner
# email: lagcleaner@gmail.com

from enum import Enum


class StatusServerCode(Enum):
    """Status Server Code Enum."""
    OK = 0
    FAILED = 1
    UNKNOWN = 2
    CONT = 3
    BREAK = 4
    DISCONNECTED = 5
    PAYLOAD_TOO_BIG = 6

    @staticmethod
    def from_str(label):
        try:
            return StatusServerCode[label]
        except (KeyError, ValueError):
            return None



# coded by lagcleaner
# email: lagcleaner@gmail.com

from tfprotocol_client.connection.protocol_client import ProtocolClient
from tfprotocol_client.misc.status_server_code import StatusServerCode
from tfprotocol_client.models.message import TfProtocolMessage


class TfProtocolKeepAliveWrapper:
    """This is an in progress class it is not finished yet, and it is intended to be used for
    some inner operations so far
    """

    def __init__(self, proto_client: ProtocolClient,) -> None:
        self._proto_client = proto_client

    @property
    def client(self):
        return self._proto_client

    def sha256_command(self, path: str) -> str:
        """Makes a sha256 hash of a file indicated by 'path'. This command is intended to apply
        a hash to a file before downloading it in order to guarantee its integrity. Be aware
        that on very large files the time to resolve the digest function could be potentially long.

        Args:
            `path` (str): The path to the target file.

        Returns:
            str: Response message.
        """
        return self.client.translate('SHA256', path).message

    def prockey_command(self) -> str:
        """Retrieves a unique key generated by the server's instance that communicates with the
        client. This unique key could be used later to identify that instance. One of these
        uses, but not the only one, is to test whether the server or even the socket communication
        line is still opened, in other words: the keepalive mechanism from the client side
        perspective.

        Returns:
            str: Response message that contains a unique key generated server side.
        """
        return self.client.translate('PROCKEY').message

    def keepalive_command(
        self, is_on: bool, time_connection: int, interval: int, count: int,
    ) -> bool:
        """Sets the configuration parameters for the TCP keepalive feature. This is especially
        useful for clients behind NAT boxes. If there is some idle time in the established
        connection -no data transmission- the NAT box could close or unset the connection
        without the peers knowing it. In contexts where it is predictable that an established
        connection could be 'in silent' for long periods of time, and it is possible that
        clients are behind NAT boxes, it is necessary to set the TCP keepalive packets.

        Args:
            `is_on` (bool): The first parameter of the command could be 0 or 1, meaning on or off.
            `time_connection` (int): The second parameter is the time (in seconds) the connection
                needs to remain idle before TCP starts sending keepalive probes.
            `interval` (int): The third parameter is the time (in seconds) between individual
                keepalive probes.
            `count` (int): The fourth parameter is the maximum number of keepalive probes TCP
                should send before dropping the connection.

        Returns:
            bool: Response message StatusServerCode.OK.
        """
        return self.client.translate(
            TfProtocolMessage(
                'KEEPALIVE',
                '1' if is_on else '0',
                str(time_connection),
                '|',
                str(interval),
                '|',
                str(count),
            )
        ).status is StatusServerCode.OK

    def date_command(self)->int:
        """Returns the number of elapsed seconds since the epoch, and arbitrary point
        in the time continuum, which is the Gregorian calendar time Jan 1 1970 00:00 UTC.

        Returns:
            int: Number of elapsed seconds since the epoch.
        """
        return int(self.client.translate('DATE').message)



# coded by lagcleaner
# email: lagcleaner@gmail.com

""" Transfer Protocol API Base implementation. """

import datetime as dt
import socket
from io import BytesIO
from multiprocessing import Condition
from typing import Callable, Union

from multipledispatch import dispatch
from tfprotocol_client.connection.protocol_client import ProtocolClient

from tfprotocol_client.security.hash_utils import hexstr, sha256_for
from tfprotocol_client.connection.codes_sender_recvr import CodesSenderRecvr
from tfprotocol_client.misc.build_utils import MessageUtils
from tfprotocol_client.misc.constants import (
    BYTE_SIZE,
    DFLT_MAX_BUFFER_SIZE,
    INT_SIZE,
    KEY_LEN_INTERVAL,
    LONG_SIZE,
    EMPTY_HANDLER,
)
from tfprotocol_client.misc.file_stat import FileStat, FileStatTypeEnum
from tfprotocol_client.misc.status_server_code import StatusServerCode
from tfprotocol_client.misc.thread import TfThread
from tfprotocol_client.models.exceptions import ErrorCode, TfException
from tfprotocol_client.models.message import TfProtocolMessage
from tfprotocol_client.models.proxy_options import ProxyOptions
from tfprotocol_client.models.putget_commands import PutGetCommandEnum
from tfprotocol_client.models.status_info import StatusInfo
from tfprotocol_client.models.transfer_state import TransferStatus
from tfprotocol_client.tfprotocol_super import TfProtocolSuper

ResponseHandler = Callable[[StatusInfo], None]
SendRecvFileHandler = Callable[[bool, str, StatusInfo, BytesIO], None]
TransferHandler = Callable[[ProtocolClient, TransferStatus], None]
TransferAsyncHandler = Callable[[CodesSenderRecvr], None]
Date = dt.date


class TfProtocol(TfProtocolSuper):
    """ Tranference Protocol API easy to use.
        `TfProtocolSuper`: The Abstract mother class of Tranference Protocol API.
    """

    # pylint: disable=super-init-not-called
    @dispatch(TfProtocolSuper, verbosity_mode=False)
    def __init__(
        self, tfprotocol: TfProtocolSuper, verbosity_mode: bool = False, **_,
    ) -> None:
        """ Constructor for Transfer Protocol class.

        Args:
            `tfprotocol` (TfProtocolSuper): A created transfer protocol instance.
            `verbosity_mode` (bool): Debug mode enabled for verbosity.
        """
        self._proto_client = tfprotocol._proto_client
        self.verbosity_mode = verbosity_mode

    # pylint: disable=function-redefined
    @dispatch(
        str,
        str,
        (str, bytes),
        str,
        int,
        proxy=ProxyOptions,
        keylen=int,
        channel_len=int,
        verbosity_mode=bool,
    )
    def __init__(
        self,
        protocol_version: str,
        public_key: str,
        client_hash: Union[str, bytes],
        address: str,
        port: int,
        proxy: ProxyOptions = None,
        keylen: int = KEY_LEN_INTERVAL[0],
        channel_len: int = DFLT_MAX_BUFFER_SIZE,
        verbosity_mode=False,
        **_,
    ) -> None:
        """ Constructor for Transfer Protocol class.

        Args:
            `protocol_version` (str, optional): The desired version of the protocol.
            `public_key` (str, optional): The previous shared rsa public key for
                the initial encryptation of the communication.
            `client_hash` (Union[str, optional, bytes]): The hash to be used by
                the server in order to test the integrity of the communication.
            `address` (str, optional): The ip/address where the protocol server is running.
            `port` (int, optional): The TCP port where the protocol is listening.
            `proxy` (ProxyOptions, optional): The proxy address for
                connect through a proxy, proxy auth. Defaults to None.
            `keylen` (int, optional): The desired length for the session key
                used to encrypt communication whit the server. Defaults to KEY_LEN_INTERVAL[0].
            `channel_len` (int, optional): The length of the channel.
                Defaults to DFLT_MAX_BUFFER_SIZE.
            `verbosity_mode` (bool): Debug mode enabled for verbosity.
        """
        super().__init__(
            protocol_version,
            public_key,
            client_hash,
            address,
            port,
            proxy,
            keylen,
            channel_len,
            verbosity_mode=verbosity_mode,
        )

    def freesp_command(self, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Retrieves the available space in the partition where the protocol folder is located.
        This command should be used only after a failure to find out if such a failure is due
        to the lack of space. It should not be used to guess how much space remains, and then
        proceed with a writing operation. The described scenario could potentially lead to a race
        condition. If both clients at the same time retrieve the free space and then write
        according to that value, one of them will fail.

        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate('FREESP'))

    def udate_command(self, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Returns the number of elapsed seconds and microseconds since the epoch -separated by
        dot-, and arbitrary point in the time continuum, which is the Gregorian calendar time
        Jan 1 1970 00:00 UTC.

        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate('UDATE'))

    def ndate_command(self, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Returns the number of elapsed seconds and nanoseconds since the epoch -separated by dot-,
        and arbitrary point in the time continuum, which is the Gregorian calendar time
        Jan 1 1970 00:00 UTC.

        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate('NDATE'))

    def echo_command(
        self, value: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Command is for debug propose. Once sent the server replays back the exact same thing
        the client sent, including the ECHO word.

        Args:
            `value` (str): The string that is going to be echoed.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('ECHO', value)).message
        )

    def mkdir_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Command creates a directory at the specified path.
        This feature allows the creation of a kind of directory that cannot be listed with
        any command. In order to access it, the name must be known in advance. The directory
        name may contain any character, except the decimal zero “0” and the path separator
        "/", and must end with an “.sd” extension. The length of the directory name must not
        exceed 255 characters including the extension. In order to create strong secured
        directories, the following technique could be used: Combine a user ́s name with a very
        strong password and apply to them the SHA256 or any other secure hash function.
        The resulting hash must be scanned to remove or change from it any occurrence of the
        decimal zero '0' -the null character- or the path separator '/'. If the hash is converted
        to hexadecimal format -which is usual- , it is guaranteed that the resulting string
        will contain only numbers and letters. If more security is needed, this process can be
        repeated recursively.

        Args:
            `path` (str): The path to the directory that is going to be created.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('MKDIR', path)))

    def del_command(self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Command deletes the specified file. This is a special command because is the only
        one who can operate in a locked directory. If it is the case, then DEL can only delete
        exactly one file: The locking file.

        Args:
            `path` (str): The target file that is going to be deleted.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('DEL', path)))

    def rmdir_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Removes a specified directory recursively.

        Args:
            `path` (str): The path of the directory that is going to be deleted.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('RMDIR', path)))

    def copy_command(
        self,
        path_from: str,
        path_to: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Copy the file indicated by the first parameter of the command to the file indicated
        by the second one.

        Args:
            `path_from` (str): The source file that is going to be copied.
            `path_to` (str): The destiny file where the file of "pathFrom" is going to be copied.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('COPY', path_from, "|", path_to))
        )

    def touch_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Creates a new file in the specified directory.

        Args:
            `path` (str): The path to the file to be created.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('TOUCH', path)))

    def date_command(
        self, response_handler: Callable[[int, StatusInfo], None] = EMPTY_HANDLER
    ):
        """Returns the number of elapsed seconds since the epoch, and arbitrary point
        in the time continuum, which is the Gregorian calendar time Jan 1 1970 00:00 UTC.

        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response = self.client.translate(TfProtocolMessage('DATE'))
        response_handler(int(response.message), response)

    def datef_command(
        self, response_handler: Callable[[Date, StatusInfo], None] = EMPTY_HANDLER
    ):
        """Returns the current date of the server in human-readable format
        “yyyy-mm-dd HH:MM:SS” UTC.

        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response = self.client.translate(TfProtocolMessage('DATEF'))
        try:
            date = dt.datetime.strptime(response.message, '%Y-%m-%d %H:%M:%S')
        except ValueError as e:
            raise TfException(exception=e)
        response_handler(date, response)

    def dtof_command(
        self,
        timestamp: Union[int, float, str],
        response_handler: Callable[[Date, StatusInfo], None] = EMPTY_HANDLER,
    ):
        """Converts date in Unix timestamp format -seconds since the epoch-
        in human-readable format “yyyy-mm-dd HH:MM:SS” UTC.

        Args:
            `timestamp` (int,float): The timestamp to be converted to human-readable string.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        if isinstance(timestamp, (int, float)):
            timestamp = str(timestamp)

        response = self.client.translate(TfProtocolMessage('DTOF', timestamp))
        try:
            date = dt.datetime.strptime(response.message, '%Y-%m-%d %H:%M:%S')
            response_handler(date, response)
        except ValueError as e:
            raise TfException(exception=e)

    def ftod_command(
        self,
        formatted_date: str,
        response_handler: Callable[[int, StatusInfo], None] = EMPTY_HANDLER,
    ):
        """Converts date in human-readable format “ yyyy-mm-dd HH:MM:SS” to its Unix
        timestamp format.

        Args:
            `formatter` (str): The date in human-readable format.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        get_time_value = dt.datetime.now() - dt.datetime(1970, 1, 1)
        timestamp = int(get_time_value.total_seconds() * 1000)

        response_handler(
            timestamp, self.client.translate(TfProtocolMessage('FTOD', formatted_date)),
        )

    def fstat_command(
        self,
        path: str,
        response_handler: Callable[[FileStat, StatusInfo], None] = EMPTY_HANDLER,
    ):
        """Returns statistics of a file or directory in the form "D | F | U FILE-SIZE
        LAST-ACCESS LAST-MODIFICATION" where D stands for directory; F stands for file;
        and U stands for unknown, only one of them is reported. The “FILE-SIZE” is
        reported in bytes in a integer of 64 bits and the number could be up to 20 digits
        long. The “LAST-ACCESS” and “LAST-MODIFICATION” are both timestamps.

        Args:
            `path` (str): The path to the target.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response = self.client.translate(TfProtocolMessage('FSTAT', path))
        if response.status is StatusServerCode.OK:
            raw_filestat = response.message.split(' ')
            response_handler(
                FileStat(
                    FileStatTypeEnum.from_char(
                        raw_filestat[0][0], dflt=FileStatTypeEnum.UNKNOWN,
                    ),
                    int(raw_filestat[1]),
                    int(raw_filestat[2]),
                    int(raw_filestat[3]),
                )
                if raw_filestat
                else None,
                response,
            )
        else:
            response_handler(
                FileStat(FileStatTypeEnum.UNKNOWN, 0, 0, 0), response,
            )

    def fupd_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Update the timestamps of the file or directory to server current time.

        Args:
            `path` (str): The path to the target which timestamp is going to be updated.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('FUPD', path)))

    def cpdir_command(
        self,
        path_from: str,
        path_to: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Copy recursively the source directory into a new created directory specified in
        the second parameter of the command.

        Args:
            `path_from` (str): The source target directory.
            `path_to` (str): The destiny target directory.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('CPDIR', path_from, '|', path_to))
        )

    def xcopy_command(
        self,
        new_name: str,
        path: str,
        pattern: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Copies the source file specified in the second parameter into every directory
        in the tree that matches “pattern” with the name “newname” specified in the first
        parameter. This is a best-effort command instead of super-reliable one. If any one of
        the destinations are locked or the file exists instead of return afailed status code,
        XCOPY skips it silently. The return status of this command are relative to the source
        file/directory.

        Args:
            `new_name` (str): The new name for the copied file.
            `path` (str): The target source file.
            `pattern` (str): The pattern that specified where to copy the target source file.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('XCOPY', new_name, path, "|", pattern)
            )
        )

    def xdel_command(
        self,
        path: str,
        file_name: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Deletes all files that match “filename” in the specified path at first parameter,
        and it does recursively start at the specified directory. This is a best-effort
        command instead of super-reliable one. If any one of the path are locked or even
        and error occurs, instead of return a failure return status, XDEL skips it silently.
        The return status of this command are relative to the path specified

        Args:
            `path` (str): The parent folder where the protocol is going to search for the files.
            `file_name` (str): The file name that you want to erase.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('XDEL', path, file_name))
        )

    def xrmdir_command(
        self,
        path: str,
        directory_name: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Deletes all directories recursively that match “directory-name” starting at
        the specified path at first parameter. This is a best-effort command instead of
        super-reliable one. If any one of the path are locked or even an error occurs,
        instead of return a failure return status, XRMDIR skips it silently. The return
        status of this command are relative to the path specified.

        Args:
            `path` (str): The top parent path, the protocol is going to search inside it
                for directoryName.
            `directory_name` (str): The name of the directories to be deleted.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('XRMDIR', path, directory_name))
        )

    def xcpdir_command(
        self,
        new_directory: str,
        path: str,
        destination_pattern: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Copy recursively source directory specified in the second parameter to every directory
        in the tree that matches “destination-directory-pattern” with the name “new-directory-name”
        specified at first parameter. This is a best-effort command instead of super-reliable one.
        If any one of the destinations are locked or the directory or file exists, instead of
        return a failure return code, XCPDIR skips it silently. The return status of this command
        are relative to the source file/directory

        Args:
            `new_directory` (str): The new name of the directory that will be created at
                specified location which is going be really the folder target specified
                at the path with a new name.
            `path` (str): The target folder that is going to be copied.
            `destination_pattern` (str): The pattern that is going to be used for selecting
                the folders.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage(
                    'XCPDIR', new_directory, path, '|', destination_pattern,
                )
            )
        )

    def lock_command(
        self, lock_filename: str, response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Specifies a filename that when it exists in a directory no operation can be done on it,
        except to delete the lock file. This allows a temporary blocking of a directory for any
        sort of operations except the one eliminating the locking file which is the command DEL
        with the path to the locking file. The above statement does not apply for commands that
        do not specify “FAILED 30” as possible return status.

        Args:
            `lock_filename` (str): The name of the file who locks directories.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('LOCK', lock_filename))
        )

    def sendfile_command(
        self,
        is_overriten: bool,
        path: str,
        stream: BytesIO,
        handler: SendRecvFileHandler = EMPTY_HANDLER,
    ):
        """Sends a file to the server.

        Args:
            `is_overriten` (bool):  The first parameter could be either “0” for false or “1”
                for true. If “/path/to/filename” it exists in the server, this flag indicates
                whether it should be overwritten or not.
            `path` (str): The path in the server where the file be stored.
            `stream` (BytesIO): The data of the file to be sent.
            `handler` (SendRecvFileHandler): The function to handle receiving states.
        """
        response = self.client.translate(
            TfProtocolMessage('SNDFILE', '1' if (is_overriten) else '0', path),
        )
        handler(
            is_overriten, path, response, stream,
        )
        while True:
            payload = stream.read(
                self.len_channel - len(StatusServerCode.CONT.name) - 1
            )
            if (
                payload
                and len(payload) - len(StatusServerCode.CONT.name) - 1
                > self.len_channel
            ):
                # PAYLOAD_TOO_BIG [CASE]
                handler(is_overriten, path, StatusInfo.parse('PAYLOAD_TOO_BIG'), stream)
                break

            # SEND DATA IF EXIST
            if payload:
                # CONNECTION BLOCKED WHEN THE PAYLOAD SENT IS THE MAXIMUM POSSIBLE,
                # USED FOR TINY FILES ONLY
                response = self.client.translate(TfProtocolMessage('CONT', payload))
                handler(is_overriten, path, response, stream)
            else:
                break

            # BREAK IF BAD RESPONSE RECEIVED
            if response.status != StatusServerCode.CONT:
                break
        handler(is_overriten, path, self.client.translate('OK'), payload)

    def rcvfile_command(
        self,
        delete_after: bool,
        path: str,
        sink: BytesIO = None,
        handler: SendRecvFileHandler = EMPTY_HANDLER,
    ):
        """Receives a file from the server.

        Args:
            `delete_after` (bool): The first parameter could be either “0” for false or “1” for
                true and tells the server whether the file must be deleted after successfully
                received by the client.
            `path` (str): The path to the file in the server to be retrieved.
            `sink` (BytesIO): The sink to send data in byte mode.
            `handler` (SendRecvFileHandler): The function to handle receiving states.
        """
        response = self.client.translate(
            TfProtocolMessage('RCVFILE', '1' if delete_after else '0', path)
        )
        handler(delete_after, path, response, sink)

        while True:
            response = self.client.translate(TfProtocolMessage('CONT'))
            handler(delete_after, path, response, sink)
            if response.status != StatusServerCode.CONT:
                break

    def ls_command(self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Command list the directory entries for the indicated path, if the argument is missing,
        it lists the root directory of the protocol daemon. The return value of this command
        is a file with the listed content. In fact, it is like issuing the command RCVFILE to
        a temporary file with the listed content of the directory.The file returned by LS has
        the following syntax.

        F | D | U /path/to/file-or-directory
        The F stands for “file”; the
        D for “directory” and the U for “unknown”.

        Args:
            `path` (str): The path to the folder to be listed.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response: StatusInfo = self.client.translate(TfProtocolMessage('LS', path))
        response_handler(response)

        while True:
            response = self.client.translate(TfProtocolMessage('CONT'))
            response_handler(response)
            if response.status != StatusServerCode.CONT:
                break

    def lsr_command(self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Command list the directory entries for the indicated path, if the argument is missing,
        it lists the root directory of the protocol daemon. The return value of this command
        is a file with the listed content. In fact, it is like issuing the command RCVFILE to
        a temporary file with the listed content of the directory.The file returned by LS has
        the following syntax.F | D | U /path/to/file-or-directoryThe F stands for “file”; the
        D for “directory” and the U for “unknown”.

        Args:
            `path` (str): The path to the folder to be listed.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response: StatusInfo = self.client.translate(TfProtocolMessage('LSR', path))
        response_handler(response)

        while True:
            response = self.client.translate(TfProtocolMessage('CONT'))
            response_handler(response)
            if response.status != StatusServerCode.CONT:
                break

    def renam_command(
        self,
        path_dir: str,
        dir_newname: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Renames the file or directory specified at first parameter into the name specified at
        second parameter. RENAM operates atomically; there is no instant at which “newname” is
        non-existent between the operation's steps if “newname” already exists. If a system
        crash occurs, it is possible for both names “oldname” and “newname” to still exist,
        but “newname” will be intact.RENAM has some restrictions to operate.1) “oldname” it
        must exist.2) If “newname” is a directory must be empty.3) If “oldname” is a directory
        then “newname” must not exist or it has to be an empty directory.4) The “newname”
        must not specify a subdirectory of the directory “oldname” which is being renamed.

        Args:
            `path_dir` (str): The target to be renamed.
            `dir_newname` (str): The new name for this target directory.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('RENAM', path_dir, '|', dir_newname)
            )
        )

    def keepalive_command(
        self,
        is_on: bool,
        time_connection: int,
        interval: int,
        count: int,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Sets the configuration parameters for the TCP keepalive feature. This is especially
        useful for clients behind NAT boxes. If there is some idle time in the established
        connection -no data transmission- the NAT box could close or unset the connection
        without the peers knowing it. In contexts where it is predictable that an established
        connection could be 'in silent' for long periods of time, and it is possible that
        clients are behind NAT boxes, it is necessary to set the TCP keepalive packets.

        Args:
            `is_on` (bool): The first parameter of the command could be 0 or 1, meaning on or off.
            `time_connection` (int): The second parameter is the time (in seconds) the connection
                needs to remain idle before TCP starts sending keepalive probes.
            `interval` (int): The third parameter is the time (in seconds) between individual
                keepalive probes.
            `count` (int): The fourth parameter is the maximum number of keepalive probes TCP
                should send before dropping the connection.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage(
                    'KEEPALIVE',
                    '1' if is_on else '0',
                    time_connection,
                    '|',
                    interval,
                    '|',
                    count,
                )
            )
        )

    def login_command(
        self, user: str, passw: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Login into system this is required for many others commands.

        Args:
            `user` (str): The username for the login action.
            `passw` (str): The password for the login action.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('LOGIN', user, passw)))

    def chmod_command(
        self,
        path_file: str,
        octal_mode: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Change the access modifier to a new one.

        Args:
            `path_file` (str): The location to the file to be modified.
            `octal_mode` (str): The octal modes stands for 000 or 777 etc in octal representation
                000 means not permission for owner not permission for user and nor for group
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('CHMOD', path_file, octal_mode))
        )

    def chown_command(
        self,
        path_file: str,
        user: str,
        group: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Change the owner of a file to a new one .This commands is similar to chown command
        at unix systems.

        Args:
            `path_file` (str): The location to the file.
            `user` (str): The new username who is going to owns the file.
            `group` (str): The new group for the file.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('CHOWN', path_file, user, group))
        )

    def putcan_command(
        self,
        data_stream: BytesIO,
        path_file: str,
        offset: int,
        buffer_size: int,
        canpt: int,
        response_handler: ResponseHandler = EMPTY_HANDLER,
        transfer_handler: TransferHandler = EMPTY_HANDLER,
    ):
        """Upload a file to the server.

        Args:
            `data_stream` (BytesIO): Input stream where you can read the information to
                be uploaded. Can be any object that inherits from BinaryIO in read mode.
            `path_file` (str): Path to server file to upload, if not exists creates it.
            `offset` (int): The pointer position from where to start writing to the
                server file.
            `buffer_size` (int): The size proposed by the client for the buffer, the buffer
                definitive will be sent by the server in case of OK.
            `canpt` (int): Cancellation Points, determines the cancellation points where the
                user, it will have to read from the server to know if it continues or not.
            `response_handler` (ResponseHandler): The function to handle the command response.
            `transference_handler` (TransferenceHandler): The function to handle the transference cancelations.
        """
        offset, canpt = max(0, offset), max(0, canpt)

        # SEND INITIAL OPTIONS
        response = self.client.translate(
            TfProtocolMessage('PUTCAN ', path_file, separate_by_spaces=False)
            .add(' ')
            .add(offset, size=LONG_SIZE, signed=False)
            .add(buffer_size, size=LONG_SIZE, signed=True)
            .add(canpt, size=LONG_SIZE, signed=False)
        )
        response_handler(response)

        # INITIALIZE TRANSFER VARIABLES
        transfer_status = TransferStatus()
        header_size = LONG_SIZE
        server_buffer_size = buffer_size
        transfer_status.command = PutGetCommandEnum.HPFCONT
        if response is None or response.code != 0:
            return
        server_buffer_size = MessageUtils.decode_int(response.payload, signed=True)
        i = 0
        while True:
            if canpt > 0 and i == canpt:
                i = 0
                cur_header = self.client.just_recv_int(size=header_size, signed=True)
                transfer_status.dummy_state = True
                if cur_header == PutGetCommandEnum.HPFCANCEL.value:
                    transfer_status.command = PutGetCommandEnum.HPFCANCEL.value
                    transfer_handler(self.client, transfer_status)
                    break
                elif cur_header == PutGetCommandEnum.HPFCONT.value:
                    transfer_status.command = PutGetCommandEnum.HPFCONT.value
                    transfer_handler(self.client, transfer_status)
                else:
                    transfer_handler(self.client, transfer_status)
                    raise TfException(
                        status_server_code=StatusServerCode.FAILED,
                        code=ErrorCode.CAN_PUT,
                        message="Some error ocurred while trying to handle canpt",
                    )
                transfer_status.dummy_state = False
                continue
            i += 1
            # LOAD DATA TO BE SENT
            payl: bytearray = None
            try:
                payl = data_stream.read(server_buffer_size)
            except Exception as e:
                raise TfException(exception=e)

            # SEND DATA CHUNK
            if payl:
                self.client.send(payl, header_size=LONG_SIZE)
            transfer_status.dummy = len(payl)
            transfer_handler(self.client, transfer_status)

            # HANDLER SIGNAL (SEND HPFCANCEL or HPFSTOP)
            if transfer_status.dummy in (
                PutGetCommandEnum.HPFCANCEL.value,
                PutGetCommandEnum.HPFSTOP.value,
            ):
                self.client.just_send(
                    transfer_status.dummy, size=header_size, signed=True
                )

            # BREAK STOP THE CYCLE IF THERE IS NO DATA LEFT IN THE 'data_stream'
            try:
                if not payl or transfer_status.dummy == PutGetCommandEnum.HPFEND.value:
                    self.client.just_send(
                        PutGetCommandEnum.HPFEND.value, size=header_size, signed=True,
                    )
                    transfer_status.dummy = PutGetCommandEnum.HPFEND.value
                    transfer_handler(self.client, transfer_status)
                    break
            except IOError as e:
                raise TfException(exception=e)

    def getcan_command(
        self,
        data_sink: BytesIO,
        path_file: str,
        offset: int,
        buffer_size: int,
        canpt: int,
        response_handler: ResponseHandler = EMPTY_HANDLER,
        transfer_handler: TransferHandler = EMPTY_HANDLER,
    ):
        """Download a file from the server.

        Args:
            `data_sink` (BytesIO): Output sink where you can write the downloaded information.
                Can be any object that inherits from BinaryIO in write mode.
            `path_file` (str): Path to server file to be download.
            `offset` (int): The pointer position from where to start reading to the server file.
            `buffer_size` (int): The size proposed by the client for the buffer, the buffer
                definitive will be sent by the server in case of OK.
            `canpt` (int): Cancellation Points, determines the cancellation points where the
                user, it will have to read from the server to know if it continues or not.
            `response_handler` (ResponseHandler): The function to handle the command response.
            `transference_handler` (TransferenceHandler): The function to handle the transference
                cancelations.
        """
        offset, canpt = max(offset, 0), max(canpt, 0)
        # SEND INITIAL OPTIONS
        response = self.client.translate(
            TfProtocolMessage('GETCAN', path_file)
            .add(' ')
            .add(offset, size=LONG_SIZE, signed=False)
            .add(buffer_size, size=LONG_SIZE, signed=True)
            .add(canpt, size=LONG_SIZE, signed=False)
        )
        response_handler(response)

        # INITIALIZE TRANSFER VARIABLES
        transfer_status = TransferStatus()
        header_size = LONG_SIZE
        if response is None or response.code != 0:
            return
        _ = MessageUtils.decode_int(response.payload, signed=True)
        i = 0
        while True:
            if canpt > 0 and i == canpt:
                i = 0
                transfer_status.dummy = None
                transfer_status.dummy_state = True

                transfer_handler(self.client, transfer_status)
                if not transfer_status.dummy:
                    transfer_status.dummy = PutGetCommandEnum.HPFCONT.value
                self.client.just_send(
                    transfer_status.dummy, size=header_size, signed=True,
                )

                if transfer_status.dummy == PutGetCommandEnum.HPFCANCEL.value:
                    break
                transfer_status.dummy_state = False
                continue
            i += 1

            cur_header = self.client.just_recv_int(size=header_size, signed=True)
            if cur_header in (
                PutGetCommandEnum.HPFCANCEL.value,
                PutGetCommandEnum.HPFEND.value,
            ):
                # CANCELATION FROM SERVER OR END OF FILE REACHED
                transfer_status.dummy = cur_header

                # handle hpffin command exchange at the end of getcan command
                transfer_handler(
                    self.client, transfer_status,
                )
                break
            pyld = self.client.just_recv(size=cur_header)
            try:
                data_sink.write(pyld)
            except IOError as e:
                raise TfException(exception=e)
            transfer_status.dummy = len(pyld)
            transfer_handler(self.client, transfer_status)

    def end_command(self):
        """Command is sent by the client to the server in order to terminate the TCP connection."""
        self.client.send('END')

    def sha256_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Makes a sha256 hash of a file indicated by 'path'. This command is intended to apply
        a hash to a file before downloading it in order to guarantee its integrity. Be aware
        that on very large files the time to resolve the digest function could be potentially long.

        Args:
            `path` (str): The path to the target file.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('SHA256', path)))

    def prockey_command(self, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Retrieves a unique key generated by the server's instance that communicates with the
        client. This unique key could be used later to identify that instance. One of these
        uses, but not the only one, is to test whether the server or even the socket communication
        line is still opened, in other words: the keepalive mechanism from the client side
        perspective.
        
        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate('PROCKEY'))

    def injail_command(
        self,
        secure_token: str,
        pathtojail_directory: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """In-jails the TFProtocol daemon in the directory specified by the second parameter.
        This is achieved only if “secure_token”, specified as first parameter, is permitted
        for that directory.

        Args:
            `secure_token` (str): This token allows you to change the jail folder to isolate your
                tfprotocol instance
            `pathtojail_directory` (str): The jail folder where tfprotocol will be executing
                (Use only if your secure token has access to it).
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('INJAIL', secure_token, '|', pathtojail_directory)
            )
        )

    def get_command(
        self,
        data_sink: BytesIO,
        path_file: str,
        offset: int,
        buffer_size: int,
        response_handler: ResponseHandler = EMPTY_HANDLER,
        transfer_handler: TransferAsyncHandler = EMPTY_HANDLER,
    ):
        """Get Command download data from server asynchronously to any OutputStream the
        user wants...

        Args:
            `data_sink` (BytesIO): It is the stream where data will resides after downloading,
                could be in memory or in disk etc.
            `path_file` (str): The path IN THE SERVER where the data is located at.
            `offset` (int): The offset is used for when we are reading the data start reading
                since the offset position for example if the data is 100 bytes long then if offset
                is 20 I will read since 20 to 100 skiping first 20 bytes. This is useful for when
                we stop an operation and wanted to restart it again later.
            `buffer_size` (int): Is the size of the buffer, while bigger faster will be the
                communication, of course server wont give you always the amount you request,
                because may be that server's bandwidth is getting filled.
            `response_handler` (ResponseHandler): The function to handle the command response.
            `transfer_handler` (TransferAsyncHandler): The function to handle the
                transference cancelations from both client and server asynchronously.
        """
        response = self.client.translate(
            TfProtocolMessage('GET', path_file)
            .add(' ')
            .add(offset, size=LONG_SIZE, signed=False)
            .add(buffer_size, size=LONG_SIZE, signed=True)
        )
        if response.status is not StatusServerCode.OK:
            response_handler(response)
            return
        response.code = MessageUtils.decode_int(response.payload, signed=True)
        response_handler(response)

        # SEEK TO THE OFFSET POSX
        try:
            data_sink.seek(offset)
        except IOError as e:
            raise TfException(exception=e)

        # SET UP SHARED VARIABLES
        cond_lock = Condition()
        code_sr = CodesSenderRecvr(self.client)
        t_handler: TfThread
        t_command: TfThread
        try:
            t_handler = TfThread(
                transfer_handler, cond_lock=Condition(), args=(code_sr,),
            )
            t_command = TfThread(
                self.__get_command_t,
                cond_lock=cond_lock,
                args=(data_sink, code_sr, response_handler),
            )
        except Exception as e:
            raise TfException(
                exception=e,
                message='FATAL ERROR!!! incorrect callback found, reinstall module...',
            )
        t_handler.start()
        t_command.start()

        with cond_lock:
            while t_command.is_alive() or t_handler.is_alive():
                cond_lock.wait(0.5)
                try:
                    if not t_handler.is_alive() and code_sr.sending_signal:
                        code_sr.send_get(PutGetCommandEnum.HPFFIN.value)
                except InterruptedError as e:
                    raise TfException(exception=e)

        t_command.join()
        t_handler.join()

        # FINAL HANDSHAKE
        if code_sr.last_command is not PutGetCommandEnum.HPFFIN.value:
            self.client.just_send(
                PutGetCommandEnum.HPFFIN.value, size=LONG_SIZE, signed=True
            )
        if code_sr.last_header is not PutGetCommandEnum.HPFFIN.value:
            self.client.just_recv_int(size=LONG_SIZE, signed=True)
        response_handler(
            StatusInfo(status=StatusServerCode.OK, code=PutGetCommandEnum.HPFFIN.value)
        )

    def __get_command_t(
        self,
        data_sink: BytesIO,
        code_sr: CodesSenderRecvr,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """This command is intended to be used ONLY by the getCommand function not by
        user DO NOT CALL THIS METHOD DIRECTLY.

        Args:
            `data_sink` (BytesIO): It is the sink where data will resides after downloading,
                could be in memory or in disk etc.
            `code_sr` (CodesSenderRecvr): Stateful codes sender and receiver helper.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        header_size = LONG_SIZE
        while True:
            try:
                header = self.client.just_recv_int(size=header_size, signed=True)
                if (
                    not code_sr.sending_signal
                    or header <= PutGetCommandEnum.HPFEND.value
                ):
                    status = StatusInfo(StatusServerCode.OK)
                    status.code = header
                    response_handler(status)
                if header <= PutGetCommandEnum.HPFEND.value:
                    code_sr.sending_signal = False
                    return
                chunk = self.client.just_recv(size=header)
                if not code_sr.sending_signal:
                    data_sink.write(chunk)
            except IOError as e:
                raise TfException(exception=e)

    def put_command(
        self,
        data_stream: BytesIO,
        path_file: str,
        offset: int,
        buffer_size: int,
        response_handler: ResponseHandler = EMPTY_HANDLER,
        transfer_handler: TransferAsyncHandler = EMPTY_HANDLER,
    ):
        """Upload a stream of data allowing to cancel at any moment. Asynchronously

        Args:
            `data_stream` (BytesIO): It is the stream where data resides in disk or memory,
                to be sent.
            `path_file` (str): The path IN THE SERVER where the data is located at.
            `offset` (int): The offset is used for when we are reading the data start reading
                since the offset position for example if the data is 100 bytes long then if
                offset is 20 I will read since 20 to 100 skiping first 20 bytes. This is useful
                for when we stop an operation and wanted to restart it again later.
            `buff_size` (int): Is the size of the buffer, while bigger faster will be the
                communication, of course server wont give you always the amount you request,
                because may be that server's bandwidth is getting filled.
            `response_handler` (ResponseHandler): The function to handle the command response.
            `transfer_handler` (TransferAsyncHandler): The function to handle the
                transference cancelations from both client and server asynchronously.
        """
        response = self.client.translate(
            TfProtocolMessage('PUT', path_file)
            .add(' ')
            .add(offset, size=LONG_SIZE, signed=False)
            .add(buffer_size, size=LONG_SIZE, signed=True)
        )
        if response.status is not StatusServerCode.OK:
            response_handler(response)
            return
        response.code = MessageUtils.decode_int(response.payload, signed=True)
        response_handler(response)

        # SEEK TO THE OFFSET POSX
        try:
            data_stream.seek(offset)
        except IOError as e:
            raise TfException(exception=e)

        # SET UP SHARED VARIABLES
        cond_lock = Condition()
        code_sr = CodesSenderRecvr(self.client)
        t_command: TfThread
        try:
            t_command = TfThread(
                self.__put_command_t,
                cond_lock=cond_lock,
                args=(data_stream, response.code, code_sr),
            )
        except Exception as e:
            raise TfException(
                exception=e,
                message='FATAL ERROR!!! incorrect callback found, reinstall module...',
            )
        # RUN THREAD
        t_command.start()
        # WHILE HEADERS NOT AN END-OF-FILE OR AN ERROR CONTINUES

        while True:
            code_sr.last_header = self.client.just_recv_int(size=LONG_SIZE, signed=True)
            if self.verbosity_mode:
                print('SERVER-HEADER:-', code_sr.last_header)
            if code_sr.last_header <= 0:
                code_sr.recveing_signal = True
                break

        t_command.join()

        # FINAL HANDSHAKE
        if code_sr.last_header != PutGetCommandEnum.HPFFIN.value:
            self.client.just_recv_int(size=LONG_SIZE, signed=True)
        code_sr.send_put(PutGetCommandEnum.HPFFIN.value)
        code_sr.block = True
        transfer_handler(code_sr)

    def __put_command_t(
        self,
        data_stream: BytesIO,
        buffer_size: int,
        code_sr: CodesSenderRecvr,
        response_handler: ResponseHandler = EMPTY_HANDLER,
        transfer_handler: TransferAsyncHandler = EMPTY_HANDLER,
    ):
        """This command is intended to be used ONLY by the put_command function not by
        user DO NOT CALL THIS METHOD DIRECTLY.

        Args:
            `data_stream` (BytesIO): It is the stream where data will take to be upload.
            `buffer_size` (int): Max buffer size of the payloads sended to the server.
            `code_sr` (CodesSenderRecvr): Stateful codes sender and receiver helper.
            `response_handler` (ResponseHandler): The function to handle the command response.
            `transfer_handler` (TransferAsyncHandler): The function to handle the
                transference cancelations from both client and server asynchronously.
        """
        while True:
            try:
                if code_sr.recveing_signal:
                    return

                readed = data_stream.read(buffer_size)
                if not readed:
                    self.client.just_send(
                        PutGetCommandEnum.HPFEND.value, size=LONG_SIZE, signed=True
                    )
                    return
                self.client.send(readed, header_size=LONG_SIZE)

                transfer_handler(code_sr)
                response_handler(StatusInfo(StatusServerCode.OK, code=len(readed)))
            except IOError as e:
                raise TfException(
                    exception=e,
                    message='Some error ocurred while trying to upload data... retry again later',
                )

    def nigma_command(
        self, keylen: int, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Change the current session key to another session key with any arbitrary size. This
        length should be a number equal or greater than 8 and multiple of 4.

        Args:
            `keylen` (int): Is the new session key length
            `response_handler` (ResponseHandler): The function to handle the command response.

        Raises:
            TfException: In case of invalid keylen.
        """
        if keylen % 4 != 0 or keylen < 8:
            raise TfException(
                code=-1,
                message='Invalid key length,'
                ' key length must be multiple of 4'
                ' and greater or equals than 8...',
            )
        response = self.client.translate(TfProtocolMessage('NIGMA', str(keylen)))
        if response.status is StatusServerCode.OK:
            hdr = self.client.just_recv_int(signed=True)
            self.client.session_key = bytes(self.client.just_recv(size=hdr))
            response_handler(
                StatusInfo(
                    status=StatusServerCode.OK,
                    message=str(bytes(self.client.session_key)),
                    payload=self.client.session_key,
                )
            )
        else:
            response_handler(response)

    def rmsd_command(
        self,
        secure_token: str,
        path_dir: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Removes a specified secure directory recursively. The first parameter is the secure
        token that allows to remove the directory specified in the second parameter. The secure
        token could be either a file or a directory named as the first parameter, inside the
        directory specified as the second parameter.

        Args:
            `secure_token` (str): This token allows you to remove the secure folder
            `path_dir` (str): The target folder to be deleted.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('RMSECDIR', secure_token, '|', path_dir)
            )
        )

    def tlb_command(self, response_handler: ResponseHandler = EMPTY_HANDLER):
        """Requests a tuple 2 ip/port from the Transfer Load Balancer pool. The clients can use
        this command -in 'advisory' way- to retrieve other servers in order to balance the
        overall traffic. By 'advisory' we mean that it's up to the clients to agree in asking for
        a server from the pool before start any further interaction.
        
        Args:
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate('TLB'))

    def sdown_command(self, path: str, data_sink: BytesIO, timeout: float):
        """Downloads the specified file in the command argument.

        Args:
            `path` (str): File path to the server file to be download.
            `data_sink` (BytesIO): Data sink open in write/append mode.
            `timeout` (float): Time out (in seconds) used to shutdown pending connection.
        """
        socket_timeout = self.client.socket.timeout
        self.client.socket.settimeout(timeout if timeout > 0 else socket_timeout)

        # SEND: SDOWN 'path/to/file'
        self.client.send(TfProtocolMessage('SDOWN', path))
        has_error = False
        header = None
        while True:
            try:
                header = self.client.just_recv_int(signed=True)
                if header > 0:
                    try:
                        data = self.client.just_recv(size=header)
                        data_sink.write(data)
                    except IOError:
                        has_error = True
                else:
                    break
            except TfException as e:
                if isinstance(e.original_exception, socket.timeout):
                    if self.verbosity_mode:
                        print('Connection timeout...')
                    break
                else:
                    raise e

        if socket_timeout > 0:
            self.client.socket.settimeout(socket_timeout)

        return header == 0 and not has_error

    def sup_command(self, path: str, data_stream: BytesIO, timeout: float):
        """Uploads the specified file in the command argument.

        Args:
            `path` (str): Path to store uploaded file.
            `data_stream` (BytesIO): Data stream open in read mode.
            `timeout` (float): Time out (in seconds) used to shutdown pending connection.
        """
        socket_timeout = self.client.socket.timeout
        self.client.socket.settimeout(timeout if timeout > 0 else socket_timeout)

        # SEND: SUP 'path/to/store/uploaded/file'
        self.client.send(TfProtocolMessage('SUP', path))
        buffer_size = self.client.max_buffer_size
        readed = b''
        header = 0
        try:
            while True:
                readed = data_stream.read(buffer_size)
                if readed:
                    self.client.send(readed, header_size=INT_SIZE)
                else:
                    break
            self.client.just_send(0, size=INT_SIZE, signed=True)
            header = self.client.just_recv_int(signed=True)
        except TfException as e:
            self.client.stop_connection()
            raise TfException(
                code=ErrorCode.ON_WRITE_OR_RECEIVE_TO_SOCKET,
                exception=e.original_exception,
                message='Socket exception',
            )
        except:  # pylint: disable=bare-except
            self.client.just_send(-1, size=INT_SIZE, signed=True)
            return False
        try:
            self.client.socket.settimeout(socket_timeout)
        except:  # pylint: disable=bare-except
            pass
        return header == 0

    def fsize_command(
        self, path_file: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Gets the file size -in bytes- of the specified file in the parameter.

        Args:
            `path_file` (str): Path to server file.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('FSIZE', path_file))
        size = self.client.just_recv_int(size=LONG_SIZE, signed=True)
        response_handler(
            StatusInfo(
                status=StatusServerCode.OK if size >= 0 else StatusServerCode.FAILED,
                code=size,
                message=str(size),
            )
        )

    def fsizels_command(
        self, path_file: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Each line in the specified file as parameter represents a path to a file. Gets the
        file size -in bytes- of each path.

        Args:
            `path_file` (str): Path to a server file with paths.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('FSIZELS', path_file))
        size: int = None
        while size != -3:
            size = self.client.just_recv_int(size=LONG_SIZE, signed=True)
            response_handler(
                StatusInfo(
                    status=StatusServerCode.CONT
                    if size != -1
                    else StatusServerCode.FAILED,
                    code=size,
                    message=str(size),
                )
            )

    def lsv2_command(
        self,
        path: str,
        path_file_to_store: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Lists the directory indicated by the first parameter. It stores all listed files in
        the file indicated by the second parameter, one file per line.

        Args:
            `path` (str): Path to be list in the server.
            `path_file_to_store` (str): Path to store the 'ls' output list.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('LSV2')
                .add(' ')
                .add(path)
                .add('@||@')
                .add(path_file_to_store),
            )
        )

    def lsrv2_command(
        self,
        path: str,
        path_file_to_store: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Lists (recursively) the directory indicated by the first parameter. It stores all
        listed files in the file indicated by the second parameter, one file per line.

        Args:
            `path` (str): Path to be list in the server.
            `path_file_to_store` (str): Path to store the 'ls' output list.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('LSRV2')
                .add(' ')
                .add(path)
                .add('@||@')
                .add(path_file_to_store),
            )
        )

    def ftype_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Returns a byte indicating the type of the file that the first parameter points to.

        Args:
            `path` (str): Path to file to get type.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('FTYPE', path))
        f_type = self.client.just_recv_int(size=BYTE_SIZE, signed=True)
        response_handler(
            StatusInfo(
                status=StatusServerCode.OK if f_type > -1 else StatusServerCode.FAILED,
                code=f_type,
                message=str(f_type),
            )
        )

    def ftypels_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Works pretty much like FTYPE except it return a byte with the type per each line in
        the file indicated by the first parameter.

        Args:
            `path` (str): Path to file containing the paths to files to get type.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('FTYPELS', path))
        f_type = 0
        while f_type != -2:
            f_type = self.client.just_recv_int(size=BYTE_SIZE, signed=True)
            response_handler(
                StatusInfo(
                    status=StatusServerCode.OK
                    if f_type != -1
                    else StatusServerCode.FAILED,
                    code=f_type,
                    message=str(f_type),
                )
            )

    def fstatls_command(
        self, path: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Returns a special structure per each line in the file indicated by the first parameter.

        Args:
            `path` (str): Path to file containing the paths to files to get stats.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('FSTATLS', path))
        _code = 0
        while _code != -2:
            fraw_stats = self.client.just_recv(size=26)
            _code, file_stat = FileStat.build_from_structure(fraw_stats)
            response_handler(file_stat)

    # Notify system
    def addntfy_command(
        self,
        token: str,
        path_file_to_listen: str = '',
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Sets a new notification wich starts name is `token`, and the directory is specified
        in the second parameter.

        Args:
            `token` (str): Start name for the notification.
            `path_file_to_listen` (str): Path to set notifications. If it is specified by an
                empty string  then the root directory of the protocol daemon is used.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('ADDNTFY')
                .add(' ')
                .add(token)
                .add(' ')
                .add(path_file_to_listen),
            )
        )

    def startnfy_command(
        self,
        interval: Union[float, str, int],
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Starts the notification system. Starts the watchdog to the list of directory specified
        by the ADDNTFY command. Once STARTNTFY is issued there is no way back to the standard mode.

        Args:
            `interval` (float, str, int): Interval to check the directories. If present,
                it's a decimal number expressed in seconds as 3.3 which means 3 seconds and
                300 miliseconds.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('STARTNTFY', str(interval)))

        while True:
            size_ms = self.client.just_recv_int(size=INT_SIZE, signed=False)
            file_notif_b = self.client.just_recv(size=size_ms)
            file_notif = None
            try:
                file_notif = MessageUtils.decode_str(file_notif_b)
            except UnicodeDecodeError:
                pass

            # Clients handles when to send back an OK or DEL response.
            response_handler(
                StatusInfo(
                    payload=file_notif_b,
                    message=file_notif,
                    status=StatusServerCode.OK,
                    code=size_ms,
                    sz=size_ms,
                ),
                send_ok=lambda: self.client.send(TfProtocolMessage('OK')),
                send_del=lambda: self.client.send(TfProtocolMessage('DEL')),
            )

    def intread_command(
        self,
        path: str,
        data_sink: BytesIO,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Integrity Read. It is intended to atomically download a file with its integrity
        checksum, for the case SHA256.

        Args:
            `path` (str): Path to file to read.
            `data_sink` (BytesIO): FileInput in write mode.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('INTREAD', path))
        ms_header = self.client.just_recv_int(size=INT_SIZE, signed=True)
        if ms_header == -1:
            # File not found or some error occurs
            response_handler(StatusInfo(status=StatusServerCode.FAILED, code=ms_header))
            return

        checksum = self.client.just_recv_str(size=66)
        file_payload = self.client.just_recv(size=ms_header)

        data_sink.write(file_payload)
        response_handler(
            StatusInfo(status=StatusServerCode.OK, code=ms_header, message=checksum)
        )

    def intwrite_command(
        self,
        path: str,
        data_stream: BytesIO,
        checksum: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Integrity Write. It is intended to atomically upload a file with its integrity
        checksum, for the case SHA256.

        Args:
            `path` (str): Path to file to write.
            `data_stream` (BytesIO): FileInput in read mode.
            `checksum` (str): Checksum of the file.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        self.client.send(TfProtocolMessage('INTWRITE', path))

        # Send 66 bytes of file hash
        self.client.just_send(checksum)

        # Send the file payload
        payload = data_stream.read()
        self.client.send(payload, header_size=INT_SIZE)

        resp_code = self.client.just_recv_int(size=INT_SIZE, signed=True)

        response_handler(
            StatusInfo(
                status=StatusServerCode.OK
                if resp_code == 0
                else StatusServerCode.FAILED,
                code=resp_code,
            )
        )

    def netlock_command(
        self,
        timeout: Union[float, str, int],
        path_lock: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Creates an advisory record lock in the file specified by the second parameter with a
        timeout in seconds indicated in the first parameter. The system guarantee that each time
        the network detects activity, the timeout will be reset.

        Args:
            `timeout` (float, str, int): Timeout to lock the network. If present, it is a
            decimal number expressed in seconds as 3.3 which means 3 seconds and 300 miliseconds.
            `path_lock` (str): Path to file to lock.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('NETLOCK', str(timeout), path_lock))
        )

    def netlocktry_command(
        self,
        timeout: Union[float, str, int],
        path_lock: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Creates an advisory record lock in the file specified by the second parameter with a
        timeout in seconds indicated in the first parameter. The system guarantee that each time
        the network detects activity, the timeout will be reset. (Works exactly like NETLOCK but
        instead of blocking until obtain the lock, it will return the error FAILED 54.)

        Args:
            `timeout` (float, str, int): Timeout to lock the network. If present, it is a
            decimal number expressed in seconds as 3.3 which means 3 seconds and 300 miliseconds.
            `path_lock` (str): Path to file to lock.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('NETLOCK_TRY', str(timeout), path_lock)
            )
        )

    def netunlock_command(
        self, lock_id: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Removes the advisory record lock in the file specified by the second parameter.

        Args:
            `path_lock` (str): Path to file to unlock.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(self.client.translate(TfProtocolMessage('NETUNLOCK', lock_id)))

    def netmutacqtry_command(
        self,
        path_mutex: str,
        token: Union[bytes, str],
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Tries to acquire the ownership of a synchronization object to enter a critical section.
        Who acquires the mutex 'token' must release it explicitly by calling NETMUTREL command,
        otherwise the mutex will be considered as acquire forever.

        Args:
            `path_mutex` (str): Path to mutex.
            `token` (bytes,str): Token to acquire the mutex.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('NETMUTACQ_TRY', path_mutex, token))
        )

    def netmutrel_command(
        self,
        path_mutex: str,
        token: Union[bytes, str],
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Releases the ownership of a synchronization object to enter a critical section.

        Args:
            `path_mutex` (str): Path to mutex.
            `token` (bytes,str): Token to release the mutex.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('NETMUTREL', path_mutex, token))
        )

    def setfsid_command(
        self,
        secure_file_sys_identity: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """sets the process identity that will be used in any operations that involves the
        TFProtocol Secure Filesystem.

        Args:
            `secure_file_sys_identity` (str): File system identity.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('SETFSID', secure_file_sys_identity)
            )
        )

    def setfsperm_command(
        self,
        secid: str,
        permission_mask: int,
        path_secref_dir: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """sets the permissions of a file or directory.

        Args:
            `secid` (str): Security file system identity, can be '' to set "other's" permissions
                (last resort resource).
            `permission_mask` (int): Permission mask.
            `path_secref_dir` (str): Path to directory.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage(
                    'SETFSPERM', f'{secid}:{permission_mask}', path_secref_dir
                )
            )
        )

    def remfsperm_command(
        self,
        secid: str,
        path_secref_dir: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Removes the secID indicated by the first parameter from the secure filesystem
        directory indicated by the second parameter.

        Args:
            `secid` (str): Security file system identity, can be '' to remove the "other's"
                permissions.
            `path_secref_dir` (str): Path to directory.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('REMFSPERM', secid, path_secref_dir)
            )
        )

    def getfsperm_command(
        self,
        secid: str,
        path_secref_dir: str,
        response_handler: ResponseHandler = EMPTY_HANDLER,
    ):
        """Gets the secID's permissions of the specified directory in the second parameter.


        Args:
            `secid` (str): Security file system identity.
            `path_secref_dir` (str): Path to directory.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(
                TfProtocolMessage('GETFSPERM', secid, path_secref_dir)
            )
        )

    def issecfs_command(
        self, path_to_dir: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Checks if 'path/to/directory' belongs to the TFProtocol Secure Filesystem domain.

        Args:
            `path_to_dir` (str): Path to directory.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('ISSECFS', path_to_dir))
        )

    def locksys_command(
        self, path_in_lock: str, response_handler: ResponseHandler = EMPTY_HANDLER
    ):
        """Locks TFProtocol in the specified directory. Once the daemon is locked in a directory,
        there isn't way back.

        Args:
            `path_in_lock` (str): Path to lock.
            `response_handler` (ResponseHandler): The function to handle the command response.
        """
        response_handler(
            self.client.translate(TfProtocolMessage('LOCKSYS', path_in_lock))
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tfprotocol_client',
 'tfprotocol_client.connection',
 'tfprotocol_client.misc',
 'tfprotocol_client.models',
 'tfprotocol_client.security']

package_data = \
{'': ['*']}

install_requires = \
['multipledispatch>=0.6.0,<0.7.0', 'pycryptodome>=3.15.0,<4.0.0']

setup_kwargs = {
    'name': 'tfprotocol-client',
    'version': '1.0.3',
    'description': 'Transfer Protocol client implemented in python.',
    'long_description': "================================================\nTFProtocol Client Implemented in Python :snake:\n================================================\n\n.. image:: https://img.shields.io/badge/Maintained%3F-yes-green.svg\n    :target: https://GitHub.com/lagcleaner/tfprotocol_client_py/graphs/commit-activity\n\n.. image:: https://img.shields.io/github/issues/lagcleaner/tfprotocol_client_py.svg\n    :target: https://GitHub.com/lagcleaner/tfprotocol_client_py/issues/\n.. image:: https://img.shields.io/github/issues-closed/lagcleaner/tfprotocol_client_py.svg\n    :target: https://GitHub.com/lagcleaner/tfprotocol_client_py/issues?q=is%3Aissue+is%3Aclosed\n\n----------------\nIntroduction \n----------------\n\nThe especifications for the *Transference Protocol* is available in this `repository\n<https://github.com/GoDjango-Development/TFProtocol/blob/main/doc/>`_.\n\n\n---------------------------\nInstallation :floppy_disk:\n---------------------------\n\n.. image:: https://img.shields.io/pypi/v/tfprotocol-client.svg\n    :target: https://pypi.org/project/tfprotocol-client/\n\nThe package is available at `pypi <https://pypi.org/project/tfprotocol-client/>`_, to be installed from **pip** with the\nnext command:\n\n.. code-block:: bash\n\n    pip install tfprotocol_client\n\n-------------------------\nA Simple Example :memo:\n-------------------------\n\nTo use the *Transference Protocol* through this library, you must create an instance of\n*TfProtocol* with the specified parameters and have an online server to connect to.\n\n.. code-block:: python\n\n    from tfprotocol_client.misc.constants import RESPONSE_LOGGER\n    from tfprotocol_client.tfprotocol import TfProtocol\n\n    ADDRESS = 'tfproto.expresscuba.com'\n    PORT = 10345\n    clienthash = '<clienthash>'\n    publickey = '<publickey>'\n\n    proto = TfProtocol('0.0', publickey, clienthash, ADDRESS, PORT)\n    proto.connect()\n    proto.echo_command('Hello World', response_handler=RESPONSE_LOGGER)\n    proto.disconnect()\n\n\n---------------------------\nFor Contributors :wrench:\n---------------------------\n\n.. image:: https://img.shields.io/github/contributors/lagcleaner/tfprotocol_client_py.svg\n    :target: https://GitHub.com/lagcleaner/tfprotocol_client_py/graphs/contributors/\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nDevelopment Environment Installation :computer:\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nTo set up the development environment, all you need as a prerequisite is to have python Python\n2.7  or 3.5+ and `poetry <https://python-poetry.org/>`_ installed. If you need to install poetry\nfollow `these steps <https://python-poetry.org/docs/#installation>`_ and come back. \n\nWith this in mind, to install the necessary dependencies and create a python environment for\nthis project, proceed to run the following command in the root directory of the project.\n\n.. code-block:: bash\n\n    poetry install\n\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\nProject Structure :open_file_folder:\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nThis library is made up of 4 folders and the particular implementations of the ``TfProtocolSuper``\nclass, the folders are structured as follows:\n\n- **connection**: where all socket and low-level communication is located.\n- **models** where the complex objects used all over the package are defined.\n- **security** where is implemented the methods and classes to encrypt and decrypt the messages for communication and also the utils for do the hashing stuff where is needed.\n- **misc** folder to hold all utils and not related to any other folder concept.\n\nHere the visual schema for all the classes and his relations with others:\n\n.. image:: ./doc/statics/classes.png\n    :alt: class relations\n    :align: center\n\n^^^^^^^^^^^^^^^^^^^^\nPublishing :rocket:\n^^^^^^^^^^^^^^^^^^^^\n\nTo publish the package you need to run the following command in the root directory of the package:\n\n.. code-block:: bash\n\n    poetry publish\n\n.. image:: https://img.shields.io/badge/Ask%20me-anything-1abc9c.svg\n    :target: mailto://lagcleaner@gmail.com\n",
    'author': 'Leonel Garcia',
    'author_email': 'lagcleaner@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lagcleaner/tfprotocol_client_py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



# coded by lagcleaner
# email: lagcleaner@gmail.com

from threading import Condition, Thread
from typing import Any, Callable, Iterable, Mapping
from tfprotocol_client.models.exceptions import TfException


class TfThread(Thread):
    """Represents a thread control, for multithreaded tasks"""

    HANDLED_THREAD = None

    def __init__(
        self,
        target: Callable[..., Any],
        quite_errors: bool = False,
        cond_lock: Condition = None,
        name: str = None,
        args: Iterable[Any] = None,
        kwargs: Mapping[str, Any] = None,
        *,
        daemon: bool = ...,
    ) -> None:
        """Preparation for the thread execution.

        Args:
            `target` (Callable[..., Any]): The method who is going to be used asynchronous.
            `quite_errors` (bool, optional): Raise silently exceptions.
            `cond_lock` (Conditional, optional): The mutex that is going to be informed when the method
                execution reach its end. Defaults to None.
            `name` (str, optional): Thread name. Defaults to ....
            `args` (Iterable[Any], optional): Postitional arguments to be passed to method
                call, must be set at the exactly order required by the method.
            `kwargs` (Mapping[str, Any], optional): Keyword arguments to be passed to method
                call.
            `daemon` (bool, optional): Demonize the thread.
        """
        super().__init__(name=name, args=args, kwargs=kwargs, daemon=daemon)
        self.conditional_lock = cond_lock if cond_lock else Condition()
        self._quite_errors = quite_errors
        self._method = target

    def setup_args(self, args: Iterable[Any] = ..., kwargs: Mapping[str, Any] = ...):
        """Set up arguments for the method.

        Args:
            args (Iterable[Any], optional): Postitional arguments to be passed to method
                call, must be set at the exactly order required by the method.
            kwargs (Mapping[str, Any], optional): Keyword arguments to be passed to method
                call.
        """
        self._args = args
        self._kwargs = kwargs

    def run(self) -> None:
        """Method representing the thread's activity.

        Raises:
            `TfException`: _description_
        """
        with self.conditional_lock:
            try:
                self._method(*self._args, **self._kwargs)
            except Exception as e:  # pylint: disable=broad-except
                if not self._quite_errors:
                    raise TfException(
                        exception=e,
                        message=f'{self.getName()}:\n {self._method.__name__}()',
                    )
            finally:
                self.conditional_lock.notify_all()
        # self._stop()
        
    def interrupt(self):
        self._quite_errors = True
        # self._stop()
    
    def set_handled_t(self):
        self.__class__.HANDLED_THREAD = self



# coded by lagcleaner
# email: lagcleaner@gmail.com

import struct
from typing import Final, Optional
from multipledispatch import dispatch
from tfprotocol_client.misc.constants import (
    BYTE_SIZE,
    DFLT_HEADER_SIZE,
    ENDIANESS,
    ENDIANESS_NAME,
    INT_SIZE,
    SHORT_SIZE,
    STRING_ENCODING,
)


class MessageUtils:
    """Message utils for comunication with transfer protocol"""

    def __init__(self, max_buffer_size: int) -> None:
        self.max_buffer_size: Final[int] = max_buffer_size

    @staticmethod
    @dispatch(str)
    def encode_value(value: str, **_) -> bytes:
        encoded_str = value.encode(STRING_ENCODING)
        return struct.pack(f'{ENDIANESS}{len(encoded_str)}s', encoded_str)

    @staticmethod
    @dispatch(int)
    # pylint: disable=function-redefined
    def encode_value(
        value: int, size: int = DFLT_HEADER_SIZE, signed=False, **_
    ) -> bytes:
        value_size = value.bit_length() // 8 + (value.bit_length() % 8 != 0)
        if value_size <= BYTE_SIZE and size <= BYTE_SIZE:
            # 1-8 bit // 1 byte
            return struct.pack(f'{ENDIANESS}{"b" if signed else "B"}', value)
        if value_size <= SHORT_SIZE and size <= SHORT_SIZE:
            # 1-16 bit // 2 bytes
            return struct.pack(f'{ENDIANESS}{"h" if signed else "H"}', value)
        if value_size <= INT_SIZE and size <= INT_SIZE:
            # 17-32 bit // 4 bytes
            return struct.pack(f'{ENDIANESS}{"i" if signed else "I"}', value)
        # 33-64 bit // 8 bytes
        return struct.pack(f'{ENDIANESS}{"q" if signed else "Q"}', value)

    @staticmethod
    @dispatch(bytes)
    # pylint: disable=function-redefined
    def encode_value(value: bytes, **_) -> bytes:
        return struct.pack(f'{ENDIANESS}{len(value)}s', value)

    @staticmethod
    @dispatch(object)
    # pylint: disable=function-redefined
    def encode_value(_, **__) -> Optional[bytes]:
        return None

    @staticmethod
    @dispatch(bool)
    # pylint: disable=function-redefined
    def encode_value(value: bool, **_) -> bytes:
        return struct.pack(f'{ENDIANESS}?', value)

    @staticmethod
    def decode_str(value: bytes, **_) -> str:
        return str(value, encoding=STRING_ENCODING)

    @staticmethod
    def decode_int(value: bytes, signed=False) -> int:
        return int.from_bytes(value, byteorder=ENDIANESS_NAME, signed=signed)

